from arcade.glui.font import BitmapFont
from arcade.glui.inputmenu import InputMenu
from arcade.glui.topmenu import GameCenterItem
from arcade.glui.menu import Menu
from arcade.glui.navigatable import Navigatable
from arcade.glui.opengl import gl, fs_emu_blending, fs_emu_texturing
from arcade.glui.render import Render
from arcade.glui.state import State
from arcade.glui.texture import Texture
from arcade.option import Option
from fsgs.context import fsgs
from fsgs.platform import PlatformHandler

GROUP_SPACING = 26
HEADING_TEXT_LEFT = 1920 - 560 + 40 + 20 - 28
ITEM_TEXT_LEFT = 1920 - 560 + 40 + 20
SIDEBAR_START_Y = 1080 - 90 - 50


# FIXME: REMOVE?
class Transition(object):
    def __init__(self):
        self.value = 0.0
        self.start = 0
        self.end = 0
        self.on_finish = None


enter_transition = Transition()
exit_transition = Transition()


def render_wall():  # brightness=1.0):
    # glClearColor(0.1, 0.1, 0.1, 1.0)
    # glClear(GL_DEPTH_BUFFER_BIT);
    Render.get().hd_perspective()
    # fs_emu_ortho();
    # fs_emu_blending(FALSE);
    # fs_emu_texturing(FALSE);
    z = -0.999

    # glPushMatrix()
    # glTranslate(0.0, 0.0, 1.0)

    # transition y-coordinate between floor and wall
    split = 361
    fs_emu_blending(False)
    fs_emu_texturing(False)
    gl.glBegin(gl.GL_QUADS)

    gl.glColor3f(39.0 / 255.0, 44.0 / 255.0, 51.0 / 255.0)
    gl.glVertex3f(0, split, z)
    gl.glVertex3f(1920, split, z)
    color = 0
    gl.glColor3f(color, color, color)
    gl.glVertex3f(1920, 1020, z)
    gl.glVertex3f(0, 1020, z)

    gl.glVertex3f(0, 1020, z)
    gl.glVertex3f(1920, 1020, z)
    gl.glVertex3f(1920, 1080, z)
    gl.glVertex3f(0, 1080, z)

    color = 0
    gl.glColor3f(color, color, color)
    gl.glVertex3f(0, 0, z)
    gl.glVertex3f(1920, 0, z)
    gl.glColor3f(20.0 / 255.0, 22.0 / 255.0, 26.0 / 255.0)
    gl.glVertex3f(1920, split, z)
    gl.glVertex3f(0, split, z)

    gl.glEnd()
    # fs_emu_texturing(True)

    # fs_emu_perspective();
    # double t_x = t0_x + (t1_x - t0_x) * g_menu_transition;
    # double t_y = t0_y + (t1_y - t0_y) * g_menu_transition;
    # double t_z = t0_z + (t1_z - t0_z) * g_menu_transition;
    # double r_a = r0_a + (r1_a - r0_a) * g_menu_transition;
    #
    # glScaled(16.0 / 9.0, 1.0, 1.0);
    # glTranslated(t_x, t_y, t_z);
    # glRotated(r_a, 0.0, 1.0, 0.0);


class GameMenu(Menu):
    def __init__(self, item):
        Menu.__init__(self)

        self.runner = None
        self.platform_handler = None
        self.last_menu_data = None

        self.temp_fix_configs(item)

        self.items.append(item)
        if self.use_game_center_item():
            self.top.left.append(GameCenterItem())
        # self.top.left.append(HomeItem())
        # self.top.left.append(MenuItem(item.title))
        self.top.set_selected_index(
            len(self.top.left) + len(self.top.right) - 1
        )

        self.context = None
        self.controller = None

        self.create_context()
        self.create_controller()
        print("GameMenu.__init__: controller is", self.controller)

        self.config_list = GameConfigList(self.runner, item)
        self.info_panel = GameInfoPanel()

        self.config_list.info_panel = self.info_panel
        # self.config_list.controller = self.controller
        self.info_panel.config_list = self.config_list

        self.navigatable = self.config_list

        # reset transitions
        enter_transition.start = State.get().time
        enter_transition.end = enter_transition.start + 0.2
        exit_transition.start = 0

    # noinspection PyMethodMayBeStatic
    def temp_fix_configs(self, item):
        # from fsgs.Database import Database
        # local_game_database = Database.get_instance()
        # game_database = fsgs.get_game_database()
        #
        # variants = local_game_database.find_game_variants_new(
        #     game_uuid=item.uuid)
        # print(variants)
        #
        # ordered_list = []
        # for variant in variants:
        #
        #     variant["like_rating"], variant["work_rating"] = \
        #         game_database.get_ratings_for_game(variant["uuid"])
        #     variant["personal_rating"], ignored = \
        #         local_game_database.get_ratings_for_game(variant["uuid"])
        #
        #     # user_rating = variant[5] or 0
        #     # global_rating = variant[3] or 0
        #     # user_rating = 0
        #     # global_rating = 0
        #
        #     variant_uuid = variant["uuid"]  # variant[2]
        #     variant_name = variant["name"]  # variant[1]
        #     variant_name = variant_name.replace("\n", " (")
        #     variant_name = variant_name.replace(" \u00b7 ", ", ")
        #     variant_name += ")"
        #     ordered_list.append(
        #         ((1 - bool(variant["have"]),
        #           1000 - variant["personal_rating"],
        #           1000 - variant["like_rating"]),
        #          (variant_uuid, variant_name, variant["database"])))
        # ordered_list.sort()
        # print("ordered variant list:")
        # for variant in ordered_list:
        #     print("-", variant[1][1])
        # item.configurations = [co[1] for co in ordered_list]
        variants = fsgs.get_ordered_game_variants(item.uuid)
        item.configurations = []
        for variant in variants:
            item.configurations.append(
                (
                    variant["uuid"],
                    variant["name"],
                    variant["database"],
                    variant["have"],
                )
            )

    def create_context(self):
        item = self.items[0]
        variant_uuid = item.configurations[0][0]
        database_name = item.configurations[0][2]
        print("\nitem:\n", item.configurations)
        print("\nitem[0]:\n", item.configurations[0])
        print("\n\nvariant_uuid =", variant_uuid, "\n\n")

        # print("configurations: ", item.configurations)
        # configs = sort_configurations(item.configurations)

        # configs = item.configurations
        # config = configs[0]
        # self.context = GameContext.create_for_game(
        #     item.platform, item.name, config)

        values = fsgs.game.set_from_variant_uuid(database_name, variant_uuid)

        # print("")
        # for key in sorted(values.keys()):
        #     print(" * {0} = {1}".format(key, values[key]))
        # print("")

        self.platform_handler = PlatformHandler.create(fsgs.game.platform.id)
        loader = self.platform_handler.get_loader(fsgs)
        fsgs.config.load(loader.load_values(values))

    def create_controller(self):
        # old_controller = self.controller
        # self.controller = GameHandler.create_for_game_context(
        #     self.context)

        self.runner = self.platform_handler.get_runner(fsgs)

        # self.controller = GameRunner()

    def recreate_controller(self):
        # self.create_controller()
        # self.config_list.set_controller(self.controller)
        print("recreate_controller, actually, just recreating game context")
        self.create_context()
        self.controller.context = self.context

    def render(self):
        # print("GameMenu.render")
        enter_transition.value = 1.0
        if enter_transition.start > 0:
            enter_transition.value = (
                State.get().time - enter_transition.start
            ) / (enter_transition.end - enter_transition.start)
            # prevent render from stopping when animating
            Render.get().dirty = True
        # transition goes from 1.0 ... 0.0
        # enter_transition.value = 1.0 - enter_transition.value
        # finished = 0
        if enter_transition.value >= 1.0:
            enter_transition.value = 1.0
            # finished = 1.0
            enter_transition.start = 0
            # enter_transition.on_finish()
            # return

        exit_transition.value = 0.0
        if exit_transition.start > 0:
            exit_transition.value = (
                State.get().time - exit_transition.start
            ) / (exit_transition.end - exit_transition.start)
            # prevent render from stopping when animating
            Render.get().dirty = True
        # transition goes from 1.0 ... 0.0
        exit_transition.value = 1.0 - exit_transition.value
        # finished = 0
        # if exit_transition.value <= 0.0:
        #     exit_transition.value = 0.0
        #     # finished = 1.0
        #     exit_transition.start = 0
        #     exit_transition.on_finish()
        #     return

        last_menu = State.get().history[-2]
        self.last_menu_data = last_menu.render()

        # self.top_menu_transition = transition

        # glClear(GL_DEPTH_BUFFER_BIT)
        # print(transition)
        # glEnable(GL_DEPTH_TEST)
        # glDepthMask(True)

        # render_wall(transition)
        # render_screen(transition)
        # if finished:
        #     # FIXME: defer to later via an event or similar?
        return None

    def render_transparent(self, data):
        # print("GameMenu.render_transparent")
        last_menu = State.get().history[-2]
        last_menu.render_transparent(self.last_menu_data)

        if exit_transition.value < 1.0:
            opacity = 1.0 - exit_transition.value
            opacity *= 1.25
            if opacity > 1.0:
                opacity = 1.0
            Render.get().hd_perspective()
            gl.glDisable(gl.GL_DEPTH_TEST)
            fs_emu_texturing(False)
            fs_emu_blending(True)
            gl.glBegin(gl.GL_QUADS)
            gl.glColor4f(0.0, 0.0, 0.0, opacity)
            gl.glVertex2f(0.0, 0.0)
            gl.glVertex2f(1920.0, 0.0)
            gl.glVertex2f(1920.0, 1020.0)
            gl.glVertex2f(0.0, 1020.0)
            gl.glEnd()
            gl.glEnable(gl.GL_DEPTH_TEST)

        alpha = enter_transition.value * 0.75

        Render.get().hd_perspective()
        gl.glDisable(gl.GL_DEPTH_TEST)
        fs_emu_texturing(False)
        fs_emu_blending(True)
        gl.glBegin(gl.GL_QUADS)
        gl.glColor4f(0.0, 0.0, 0.0, alpha)
        # Covers, left
        gl.glVertex2f(0.0, 366.0)
        gl.glVertex2f(746.0, 366.0)
        gl.glVertex2f(746.0, 1020.0)
        gl.glVertex2f(0.0, 1020.0)
        # Covers, right
        gl.glVertex2f(1920.0 - 746.0, 366.0)
        gl.glVertex2f(1920.0, 366.0)
        gl.glVertex2f(1920.0, 1020.0)
        gl.glVertex2f(1920.0 - 746.0, 1020.0)
        # Bottom
        gl.glColor4f(0.0, 0.0, 0.0, 0.50)
        gl.glVertex2f(0.0, 0.0)
        gl.glVertex2f(1920.0, 0.0)
        gl.glVertex2f(1920.0, 366.0)
        gl.glVertex2f(0.0, 366.0)
        gl.glEnd()
        gl.glEnable(gl.GL_DEPTH_TEST)

        transition = enter_transition.value
        if exit_transition.value < 1.0:
            # transition = exit_transition.value
            transition = 1.0 + 3.0 - exit_transition.value * 3.0
        # print("render, transition = ", transition)
        self.config_list.render(transition)

        if exit_transition.value <= 0.0:
            exit_transition.value = 0.0
            # finished = 1.0
            exit_transition.start = 0
            exit_transition.on_finish()

    def go_up(self):
        print("GameMenu.go_up")

    def go_down(self):
        print("GameMenu.go_down")

    def go_left(self, count=1):
        print("GameMenu.go_left")

    def go_right(self, count=1):
        print("GameMenu.go_right")

    def activate(self):
        print("GameMenu.activate")


class GameConfigList(Navigatable):
    def __init__(self, controller, game_item):
        self.index = 0
        self.runner = None
        self.game_item = game_item
        self.items = []
        self.set_controller(controller)

    def set_controller(self, controller):

        self.items = []
        self.items.append(Option.create_group("Game Control", -2.5))
        play_option = Option.create_play_option()
        self.items.append(play_option)

        self.items.append(Option.create_group("Variant", -2.5))
        config_option = Option.create_config_option()
        config_option.title = self.game_item.configurations[0][1]
        self.items.append(config_option)

        if len(self.game_item.configurations) > 1:
            self.items.append(Option.create_group("Other Variants", -2.5))
            for config in self.game_item.configurations[1:]:
                config_option = Option.create_config_option()
                config_option.title = config[1]
                self.items.append(config_option)

        # self.items = controller.options
        self.index = 0
        # set index to first non-group item
        for i, item in enumerate(self.items):
            if not item.group:
                self.index = i
                break
        self.runner = controller
        # self.controller = controller
        # print("GameConfigList.set_controller", id(self.controller))

    def go_up(self):
        old_index = self.index
        if self.index > 0:
            new_index = self.index - 1
            while new_index > 0 and self.items[new_index].group:
                new_index -= 1
        else:
            new_index = 0
        if new_index != old_index and not self.items[new_index].group:
            self.index = new_index
        else:
            Navigatable.go_up(self)

    def go_down(self):
        old_index = self.index
        if self.index < len(self.items) - 1:
            new_index = self.index + 1
            while (
                new_index < len(self.items) - 1 and self.items[new_index].group
            ):
                new_index += 1
        else:
            new_index = len(self.items) - 1
        if new_index != old_index and not self.items[new_index].group:
            self.index = new_index

    def go_left(self, count=1):
        print("GameConfigList.go_left")
        pass

    def go_right(self, count=1):
        pass

    def activate(self):
        # Hack for variants
        if self.index > 4:
            # Put chosen configuration at the top
            temp_config = self.game_item.configurations[0]
            # temp_config = self.game_item.configurations[self.index - 4 + 1]
            self.game_item.configurations[0] = self.game_item.configurations[
                self.index - 4
            ]
            self.game_item.configurations[self.index - 4] = temp_config
            # Put chosen variant item at the top
            temp = self.items[3]
            self.items[3] = self.items[self.index]
            self.items[self.index] = temp
            # Select play game option
            self.index = 1
            # Hackish, please clean up...
            print("ACTIVE CONFIGURATION:", self.game_item.configurations[0])
            State.get().current_menu.create_context()
            State.get().current_menu.create_controller()
            self.runner = State.get().current_menu.runner
            return
        result = self.items[self.index].activate()
        if result == "PLAY":

            # noinspection PyDecorator
            # @staticmethod
            def show_input():
                # print("Create input menu, controller = ", id(self.controller))
                # new_menu = InputMenu(self.game_item, self.controller)
                # State.get().history.append(new_menu)
                # # FIXME
                # from .window import set_current_menu
                # set_current_menu(new_menu)

                print("Create input menu, runner = ", id(self.runner))
                new_menu = InputMenu(self.game_item, self.runner)
                State.get().history.append(new_menu)
                # FIXME
                from arcade.glui.window import set_current_menu

                set_current_menu(new_menu)

            exit_transition.start = State.get().time
            exit_transition.end = exit_transition.start + 0.5
            exit_transition.on_finish = show_input

    def render(self, transition=1.0):
        Render.get().hd_perspective()

        # w = 560
        # h = 92
        # x = 1920 - w * transition
        # y = 1080 - 60 - h
        # y = 1080 - 60 - h - 44
        # z = -0.9

        # item_top = 1080 - 90 - 50
        # item_left = 1920 - 560 + 40

        dx = Texture.sidebar_background_shadow.w * (1.0 - transition)

        gl.glPushMatrix()
        gl.glTranslate(0.0, 0.0, 0.7)
        # glTranslated((1.0 - transition) * g_tex_sidebar_background->width,
        # 0, 0)
        # fs_emu_set_texture(g_tex_sidebar_background)
        # fs_emu_render_texture_with_size(g_tex_sidebar_background,
        #         1920 - g_tex_sidebar_background->width,
        #         0, g_tex_sidebar_background->width,
        #         g_tex_sidebar_background->height)
        # fs_emu_render_sidebar()
        fs_emu_blending(True)
        fs_emu_texturing(True)
        gl.glDepthMask(False)
        Texture.sidebar_background_shadow.render(
            1920 - Texture.sidebar_background_shadow.w * transition,
            0,
            Texture.sidebar_background_shadow.w,
            Texture.sidebar_background_shadow.h,
        )
        if transition > 1.0:
            padding = 1920
            Texture.sidebar_background.render(
                1920
                - Texture.sidebar_background_shadow.w * transition
                + Texture.sidebar_background_shadow.w,
                0,
                1920 + padding,
                Texture.sidebar_background.h,
            )
        gl.glDepthMask(True)
        gl.glPopMatrix()

        y = SIDEBAR_START_Y
        for i, item in enumerate(self.items):
            if item.group:
                if i > 0:
                    y -= GROUP_SPACING
            selected = i == self.index and State.get().navigatable == self
            # z = 0.71
            fs_emu_texturing(False)
            gl.glDisable(gl.GL_DEPTH_TEST)

            if selected:
                fg_color = [1.0, 1.0, 1.0, 1.0]
                # gl.glBegin(gl.GL_QUADS)
                #
                # gl.glColor3f(0.00, 0x99 / 0xff, 0xcc / 0xff)
                # gl.glVertex3f(x, y - 18, z)
                # gl.glVertex3f(x + w, y - 18, z)
                # gl.glVertex3f(x + w, y + h - 4, z)
                # gl.glVertex3f(x, y + h - 4, z)
                # # glColor3f(0.6, 0.6, 0.6)
                # # glVertex3f(x, y + 4, z)
                # # glVertex3f(x + w, y + 4, z)
                # # glVertex3f(x + w, y + h, z)
                # # glVertex3f(x, y + h, z)
                # gl.glEnd()
                fs_emu_blending(True)
                fs_emu_texturing(True)
                gl.glDepthMask(False)
                Texture.item_background.render(1920 - 540 + 13 + dx, y - 18)
                gl.glDepthMask(True)
            else:
                fg_color = [1.0, 1.0, 1.0, 1.0]

            # glBegin(GL_QUADS)
            # glColor3f(0.4, 0.4, 0.4)
            # glVertex3f(x, y, z)
            # glVertex3f(x + w, y, z)
            # glVertex3f(x + w, y + 4, z)
            # glVertex3f(x, y + 4, z)
            # glEnd()
            # fs_emu_texturing(True)

            if i > 0:
                fg_color[3] *= 0.35

            text = item.title

            # FIXME: REMOVE
            # fs_emu_blending(False)
            # fs_emu_blending(True)

            if item.group:
                BitmapFont.menu_font.render(
                    text,
                    HEADING_TEXT_LEFT + dx,
                    y + 14,
                    r=0.0,
                    g=0x99 / 0xFF,
                    b=0xCC / 0xFF,
                )
                x, _ = BitmapFont.menu_font.measure(text)
                x += HEADING_TEXT_LEFT + 12
                fs_emu_blending(True)
                fs_emu_texturing(True)
                # gl.glDepthMask(False)
                Texture.heading_strip.render(x + dx, y + 14, 1920 - x, 32)
            else:
                BitmapFont.menu_font.render(text, ITEM_TEXT_LEFT + dx, y + 14)
            gl.glEnable(gl.GL_DEPTH_TEST)
            # text = item.title.upper()
            # tw, th = Render.get().text(text, Font.main_path_font,
            #         x + 40, y + 43, color=fg_color)
            # text = item.subtitle.upper()
            # fg_color[3] = fg_color[3] * 0.4
            # text = item.title.upper()
            # tw, th = Render.get().text(text, Font.list_subtitle_font,
            #         x + 40, y + 18, color=fg_color)
            y -= 54

        # Gradually erase menu items when transitioning
        if transition > 1.0:
            gl.glDisable(gl.GL_DEPTH_TEST)
            # tr = (transition - 1.0) / 3.0
            alpha = max(0.0, (transition - 1.0) / 2.0)
            Texture.sidebar_background.render(
                1920 - Texture.sidebar_background_shadow.w * transition + 200,
                0,
                1920,
                Texture.sidebar_background.h,
                opacity=alpha,
            )
            gl.glEnable(gl.GL_DEPTH_TEST)


class GameInfoPanel(Navigatable):
    def __init__(self):
        self.index = 0
        self.position = 0
        self.config_list = []

    def go_up(self):
        print("GameInfoPanel.go_up")
        if self.index == 0 and self.position == 0:
            Navigatable.go_up(self)

    def go_down(self):
        print("GameInfoPanel.go_down")

    def go_left(self, count=1):
        print("GameInfoPanel.go_left")

    def go_right(self, count=1):
        State.get().navigatable = self.config_list

    def render(self):
        z = -0.999
        fs_emu_texturing(False)
        fs_emu_blending(False)
        gl.glBegin(gl.GL_QUADS)
        gl.glColor3f(0.0, 0.0, 0.0)
        # glVertex3f(0, 365, z)
        # glVertex3f(1270, 365, z)
        gl.glVertex3f(0, 400, z)
        gl.glVertex3f(1270, 400, z)
        gl.glColor3f(0.25, 0.25, 0.25)
        gl.glVertex3f(1270, 1080 - 60, z)
        gl.glVertex3f(0, 1080 - 60, z)
        gl.glEnd()
        # fs_emu_texturing(True)



import hashlib
import os
from binascii import unhexlify

from fsbc import settings
from fsgs.drivers.gamedriver import GameDriver, Emulator
from fsgs.drivers.mednafendriver import MednafenDriver
from fsgs.drivers.mess.messnesdriver import MessNesDriver
from fsgs.drivers.retroarchdriver import RetroArchDriver
from fsgs.option import Option
from fsgs.platform import Platform
from fsgs.platforms.loader import SimpleLoader

NES_PLATFORM_ID = "nes"
NES_PLATFORM_NAME = "Nintendo"
NES_MODEL_NTSC = "ntsc"
NES_MODEL_PAL = "pal"
NES_MODEL_FAMICOM = "ntsc-j"
NES_CONTROLLER_TYPE = "gamepad"
NES_CONTROLLER = {
    "type": NES_CONTROLLER_TYPE,
    "description": "NES Gamepad",
    "mapping_name": "nintendo",
}
NES_ZAPPER_CONTROLLER_TYPE = "zapper"
NES_ZAPPER_CONTROLLER = {
    "type": NES_ZAPPER_CONTROLLER_TYPE,
    "description": "NES Zapper",
    "mapping_name": "",
}
NES_ARKANOID_CONTROLLER_TYPE = "arkanoid"
NES_ARKANOID_CONTROLLER = {
    "type": NES_ARKANOID_CONTROLLER_TYPE,
    "description": "Arkanoid Paddle",
    "mapping_name": "",
}
NO_CONTROLLER_TYPE = "none"
NO_CONTROLLER = {
    "type": NO_CONTROLLER_TYPE,
    "description": "None",
    "mapping_name": "",
}

NES_PORTS = [
    {
        "description": "Port 1",
        "types": [
            NES_CONTROLLER,
            NES_ZAPPER_CONTROLLER,
            NES_ARKANOID_CONTROLLER,
            NO_CONTROLLER,
        ],
        "type_option": "nes_port_1_type",
        "device_option": "nes_port_1",
    },
    {
        "description": "Port 2",
        "types": [
            NES_CONTROLLER,
            NES_ZAPPER_CONTROLLER,
            NES_ARKANOID_CONTROLLER,
            NO_CONTROLLER,
        ],
        "type_option": "nes_port_2_type",
        "device_option": "nes_port_2",
    },
    {
        "description": "Port 3",
        "types": [NES_CONTROLLER, NO_CONTROLLER],
        "type_option": "nes_port_3_type",
        "device_option": "nes_port_3",
    },
    {
        "description": "Port 4",
        "types": [NES_CONTROLLER, NO_CONTROLLER],
        "type_option": "nes_port_4_type",
        "device_option": "nes_port_4",
    },
]


class NintendoPlatform(Platform):
    PLATFORM_NAME = NES_PLATFORM_NAME

    def driver(self, fsgc):
        # if settings.get(Option.NES_DRIVER) == "mess":
        #     return MessNesDriver(fsgc)
        # else:
        #     return NintendoMednafenDriver(fsgc)
        driver = settings.get(Option.NES_EMULATOR)

        # FIXME: nestopia-libretro? retroarch-nestopia?
        if driver in ["retroarch-nestopia", "libretro-nestopia"]:
            return NintendoRetroArchDriver(fsgc)
        elif driver in ["mame", "mess"]:
            return MessNesDriver(fsgc)
        elif driver == "higan":
            return NintendoHiganDriver(fsgc)
        return NintendoMednafenDriver(fsgc)

    def loader(self, fsgc):
        return NintendoLoader(fsgc)


class NintendoLoader(SimpleLoader):
    def load_extra(self, values):
        self.config[Option.NES_MODEL] = values["nes_model"]
        self.config[Option.NES_INES_HEADER] = values["ines_header"]
        self.config[Option.NES_PORT_1_TYPE] = values["nes_port_1_type"]
        self.config[Option.NES_PORT_2_TYPE] = values["nes_port_2_type"]
        self.config[Option.NES_PORT_3_TYPE] = values["nes_port_3_type"]
        self.config[Option.NES_PORT_4_TYPE] = values["nes_port_4_type"]

        # FIXME: Temporary
        if self.config[Option.NES_MODEL] == "famicom":
            self.config[Option.NES_MODEL] = NES_MODEL_FAMICOM


class NintendoMednafenDriver(MednafenDriver):
    PORTS = NES_PORTS

    def __init__(self, fsgs):
        super().__init__(fsgs)
        self.helper = NintendoHelper(self.options)
        # port_2 = self.options[Option.NES_PORT_2_TYPE]
        # if port_2 == NES_ARKANOID_CONTROLLER_TYPE:
        #     print("[DRIVER] NES Port 2 type:", port_2)
        #     for i, t in enumerate(self.ports[1].types):
        #         if t["type"] == NES_ARKANOID_CONTROLLER_TYPE:
        #             self.ports[1].index = i
        #             break
        port_3 = self.options[Option.NES_PORT_3_TYPE]
        port_4 = self.options[Option.NES_PORT_4_TYPE]
        if not port_3 and not port_4:
            # Remove the last two input ports - not in use.
            self.ports[2:4] = []

    def prepare(self):
        print("[DRIVER] Mednafen NES driver preparing...")
        super().prepare()
        pfx = self.mednafen_system_prefix()

        self.emulator.args.extend(
            ["-{}.input.port1".format(pfx), self.ports[0].type]
        )
        self.emulator.args.extend(
            ["-{}.input.port2".format(pfx), self.ports[1].type]
        )

        if self.helper.model() == NES_MODEL_PAL:
            self.set_model_name("Nintendo (PAL)")
            self.emulator.args.extend(["-{}.pal".format(pfx), "1"])
        else:
            if self.helper.model() == NES_MODEL_FAMICOM:
                self.set_model_name("Famicom")
            else:
                self.set_model_name("Nintendo (NTSC)")
            self.emulator.args.extend(["-{}.pal".format(pfx), "0"])
        self.emulator.args.extend(["-{}.fnscan".format(pfx), "0"])

        self.emulator.env["FSGS_ASPECT"] = "4/3"

        viewport = self.options[Option.VIEWPORT]
        if viewport == "0 0 256 240 = 0 0 256 240":
            self.emulator.env["FSGS_CROP"] = "0,0,256,240"
        # elif viewport == "0 0 256 240 = 0 8 256 224":
        #     self.emulator.env["FSGS_CROP"] = "0,8,256,224"
        elif viewport == "0 0 256 240 = 8 0 240 240":
            self.emulator.env["FSGS_CROP"] = "8,0,240,240"
        elif viewport == "0 0 256 240 = 8 8 240 224":
            self.emulator.env["FSGS_CROP"] = "8,8,240,224"
        else:
            self.emulator.env["FSGS_CROP"] = "0,8,256,224"

        self.emulator.args.append(self.helper.prepare_rom(self))

    def finish(self):
        print("[DRIVER] Mednafen NES Driver finishing...")
        super().finish()

    # def mednafen_extra_graphics_options(self):
    #     options = []
    #     if self.nes_clip_sides():
    #         options.extend(["-nes.clipsides", "1"])
    #     return options

    def mednafen_input_mapping(self, port):
        n = port + 1
        return {
            "A": "nes.input.port{}.gamepad.a".format(n),
            "B": "nes.input.port{}.gamepad.b".format(n),
            "UP": "nes.input.port{}.gamepad.up".format(n),
            "DOWN": "nes.input.port{}.gamepad.down".format(n),
            "LEFT": "nes.input.port{}.gamepad.left".format(n),
            "RIGHT": "nes.input.port{}.gamepad.right".format(n),
            "SELECT": "nes.input.port{}.gamepad.select".format(n),
            "START": "nes.input.port{}.gamepad.start".format(n),
        }

    def mednafen_rom_extensions(self):
        return [".nes"]

    def mednafen_system_prefix(self):
        return "nes"

    def game_video_par(self):
        size = self.game_video_size()
        return (4 / 3) / (size[0] / size[1])

    def game_video_size(self):
        # FIXME
        if self.is_pal():
            size = (256, 240)
        else:
            size = (256, 224)
        # if self.nes_clip_sides():
        #     size = (size[0] - 16, size[1])
        return size

    # def nes_clip_sides(self):
    #     # FIXME: Sane default? -Or enable this in a per-game config
    #     # instead? SMB3 looks better with this
    #     # return True
    #     return False

    def get_game_file(self, config_key="cartridge_slot"):
        # path = super().get_game_file()
        #
        # # FIXME: Replace and do in one go without going via super
        # input_stream = self.fsgc.file.open(path)
        # _, ext = os.path.splitext(path)
        # return self.prepare_rom_with_stream(self, input_stream, ext)
        # return self.helper.prepare_rom(self)
        return None


class NintendoRetroArchDriver(RetroArchDriver):
    PORTS = NES_PORTS

    def __init__(self, fsgc):
        super().__init__(fsgc, "nestopia_libretro", "RetroArch/Nestopia")
        self.helper = NintendoHelper(self.options)

    def prepare(self):
        super().prepare()
        with self.open_retroarch_core_options() as f:
            if self.helper.model() == NES_MODEL_PAL:
                self.set_model_name("Nintendo (PAL)")
                f.write("nestopia_favored_system = pal\n")
            elif self.helper.model() == NES_MODEL_FAMICOM:
                self.set_model_name("Famicom")
                f.write("nestopia_favored_system = famicom\n")
            else:
                self.set_model_name("Nintendo (NTSC)")
                f.write("nestopia_favored_system = ntsc\n")

            viewport = self.options[Option.VIEWPORT]
            if viewport == "0 0 256 240 = 0 0 256 240":
                overscan_h, overscan_v = "disabled", "disabled"
            # elif viewport == "0 0 256 240 = 0 8 256 224":
            #     return 256, 224
            elif viewport == "0 0 256 240 = 8 0 240 240":
                overscan_h, overscan_v = "enabled", "disabled"
            elif viewport == "0 0 256 240 = 8 8 240 224":
                overscan_h, overscan_v = "enabled", "enabled"
            else:
                overscan_h, overscan_v = "disabled", "enabled"
            f.write("nestopia_overscan_h = {}\n".format(overscan_h))
            f.write("nestopia_overscan_v = {}\n".format(overscan_v))
        self.emulator.args.append(self.helper.prepare_rom(self))

    def game_video_size(self):
        # FIXME: Account for horizontal overscan
        # FIXME: Account for vertical overscan

        viewport = self.options[Option.VIEWPORT]
        if viewport == "0 0 256 240 = 0 0 256 240":
            return 256, 240
        # elif viewport == "0 0 256 240 = 0 8 256 224":
        #     return 256, 224
        elif viewport == "0 0 256 240 = 8 0 240 240":
            return 240, 240
        elif viewport == "0 0 256 240 = 8 8 240 224":
            return 240, 224
        else:
            return 256, 224

    def retroarch_input_mapping(self, port):
        n = port + 1
        return {
            "A": "input_player{}_a".format(n),
            "B": "input_player{}_b".format(n),
            "UP": "input_player{}_up".format(n),
            "DOWN": "input_player{}_down".format(n),
            "LEFT": "input_player{}_left".format(n),
            "RIGHT": "input_player{}_right".format(n),
            "SELECT": "input_player{}_select".format(n),
            "START": "input_player{}_start".format(n),
        }

    # def window_size(self):
    #     return 256, 224


class NintendoHiganDriver(GameDriver):
    PORTS = NES_PORTS

    def __init__(self, fsgs):
        super().__init__(fsgs)
        self.emulator = Emulator("higan")
        self.emulator.allow_system_emulator = True
        self.helper = NintendoHelper(self.options)

    def prepare(self):
        if self.use_fullscreen():
            self.emulator.args.append("--fullscreen")

        # model = self.helper.nes_model()

        rom_path = self.get_game_file()
        self.helper.fix_ines_rom(rom_path)
        self.emulator.args.extend([rom_path])


class NintendoHelper:
    def __init__(self, options):
        self.options = options

    def model(self):
        model = self.options[Option.NES_MODEL]
        if model in ["", NES_MODEL_NTSC]:
            return NES_MODEL_NTSC
        elif model == NES_MODEL_PAL:
            return NES_MODEL_PAL
        elif model == NES_MODEL_FAMICOM:
            return NES_MODEL_FAMICOM
        else:
            print("[NES] Warning: Invalid model:", model)
            return NES_MODEL_NTSC

    def pal(self):
        return self.model() == NES_MODEL_PAL

    def fix_ines_rom(self, path):
        data1 = b""
        with open(path, "rb") as f:
            # Start iNES Hack -------------------------------------------------
            data = f.read(16)
            if len(data) == 16 and data.startswith(b"NES\x1a"):
                print("[DRIVER] Stripping iNES header")
            else:
                # No iNES header, include data
                data1 = data
            # End iNES Hack ---------------------------------------------------
            data2 = f.read()
        with open(path, "wb") as f:
            ines_header = self.options[Option.NES_INES_HEADER]
            if ines_header:
                assert len(ines_header) == 32
                f.write(unhexlify(ines_header))
            f.write(data1)
            f.write(data2)
        return path

    def prepare_rom(self, driver):
        file_uri = self.options[Option.CARTRIDGE_SLOT]
        input_stream = driver.fsgc.file.open(file_uri)
        _, ext = os.path.splitext(file_uri)
        return self.prepare_rom_with_stream(driver, input_stream, ext)

    def prepare_rom_with_stream(self, driver, input_stream, ext):
        # This should not be necessary for files found via the file database
        # and the online game database, but could be necessary for manually
        # loaded files.
        data = input_stream.read(16)
        if len(data) == 16 and data.startswith(b"NES\x1a"):
            print("[DRIVER] Stripping iNES header")
            data = None
        else:
            # No iNES header, include data
            pass

        sha1_obj = hashlib.sha1()

        path = driver.temp_file("rom" + ext).path
        with open(path, "wb") as f:
            header = self.options[Option.NES_INES_HEADER]
            if header:
                assert len(header) == 16 * 2
                f.write(unhexlify(header))
                sha1_obj.update(unhexlify(header))
            if data is not None:
                f.write(data)
                sha1_obj.update(data)
            while True:
                data = input_stream.read(65536)
                if not data:
                    break
                f.write(data)
                sha1_obj.update(data)
        new_path = os.path.join(
            os.path.dirname(path), sha1_obj.hexdigest()[:8].upper() + ext
        )
        os.rename(path, new_path)
        return new_path



from __future__ import absolute_import
from __future__ import print_function

import os
import os.path
import stat
import struct
from .ADFBlockDevice import ADFBlockDevice
from .HDFBlockDevice import HDFBlockDevice
from .RawBlockDevice import RawBlockDevice
from .DiskGeometry import DiskGeometry
from amitools.fs.rdb.RDisk import RDisk
import amitools.util.BlkDevTools as BlkDevTools

class BlkDevFactory:
  """the block device factory opens or creates image files suitable as a block device for file system access."""

  valid_extensions = ('.adf','.adz','.adf.gz','.hdf','.rdisk')

  TYPE_ADF = 1
  TYPE_HDF = 2
  TYPE_RDISK = 3

  def detect_type(self, img_file, fobj, options=None):
    """try to detect the type of a given img_file name"""
    # 1. take type from options
    t = self.type_from_options(options)
    if t == None:
      # 2. look in file
      t = self.type_from_contents(img_file, fobj)
      if t == None:
        # 3. from extension
        t = self.type_from_extension(img_file)
    return t

  def type_from_options(self, options):
    """look in options for type"""
    if options != None:
      if 'type' in options:
        t = options['type'].lower()
        if t in ('adf','adz'):
          return self.TYPE_ADF
        elif t == 'hdf':
          return self.TYPE_HDF
        elif t == 'rdisk':
          return self.TYPE_RDISK
    return None

  def type_from_contents(self, img_file, fobj):
    """look in first 4 bytes for type of image"""
    # load 4 bytes
    if fobj is None:
      # make sure file exists
      if not os.path.exists(img_file):
        return None
      f = open(img_file, "rb")
      hdr = f.read(4)
      f.close()
    else:
      hdr = fobj.read(4)
      fobj.seek(0,0)
    # check for 'RDISK':
    if hdr == 'RDSK':
      return self.TYPE_RDISK
    return None

  def type_from_extension(self, img_file):
    """look at file extension for type of image"""
    ext = img_file.lower()
    if ext.endswith('.adf') or ext.endswith('.adz') or ext.endswith('.adf.gz'):
      return self.TYPE_ADF
    elif ext.endswith(".hdf"):
      return self.TYPE_HDF
    elif ext.endswith(".rdsk"):
      return self.TYPE_RDISK
    else:
      return None

  def open(self, img_file, read_only=False, options=None, fobj=None):
    """open an existing image file"""
    # file base check
    if fobj is None:
      # make sure image file exists
      if not os.path.exists(img_file):
        raise IOError("image file not found")
      # is readable?
      if not os.access(img_file, os.R_OK):
        raise IOError("can't read from image file")
      # is writeable? -> no: enforce read_only
      if not os.access(img_file, os.W_OK):
        read_only = True
      # check size
      st = os.stat(img_file)
      mode = st.st_mode
      if stat.S_ISBLK(mode) or stat.S_ISCHR(mode):
        size = BlkDevTools.getblkdevsize(img_file)
      else:
        size = os.path.getsize(img_file)
      if size == 0:
        raise IOError("image file is empty")
    # fobj
    else:
      fobj.seek(0,2)
      size = fobj.tell()
      fobj.seek(0,0)
    # detect type
    t = self.detect_type(img_file, fobj, options)
    if t == None:
      raise IOError("can't detect type of image file")
    # create blkdev
    if t == self.TYPE_ADF:
      blkdev = ADFBlockDevice(img_file, read_only, fobj=fobj)
      blkdev.open()
    elif t == self.TYPE_HDF:
      # detect geometry
      geo = DiskGeometry()
      if not geo.detect(size, options):
        raise IOError("can't detect geometry of HDF image file")
      blkdev = HDFBlockDevice(img_file, read_only, fobj=fobj)
      blkdev.open(geo)
    else:
      rawdev = RawBlockDevice(img_file, read_only, fobj=fobj)
      rawdev.open()
      # create rdisk instance
      rdisk = RDisk(rawdev)
      if not rdisk.open():
        raise IOError("can't open rdisk of image file")
      # determine partition
      p = "0"
      if options != None and options.has_key('part'):
        p = str(options['part'])
      part = rdisk.find_partition_by_string(p)
      if part == None:
        raise IOError("can't find partition in image file")
      blkdev = part.create_blkdev(True) # auto_close rdisk
      blkdev.open()
    return blkdev

  def create(self, img_file, force=True, options=None, fobj=None):
    if fobj is None:
      # make sure we are allowed to overwrite existing file
      if os.path.exists(img_file):
        if not force:
          raise IOError("can't overwrite existing image file")
        # not writeable?
        if not os.access(img_file, os.W_OK):
          raise IOError("can't write image file")
    # detect type
    t = self.detect_type(img_file, fobj, options)
    if t == None:
      raise IOError("can't detect type of image file")
    if t == self.TYPE_RDISK:
      raise IOError("can't create rdisk. use rdbtool first")
    # create blkdev
    if t == self.TYPE_ADF:
      blkdev = ADFBlockDevice(img_file, fobj=fobj)
      blkdev.create()
    else:
      # determine geometry from size or chs
      geo = DiskGeometry()
      if not geo.setup(options):
        raise IOError("can't determine geometry of HDF image file")
      blkdev = HDFBlockDevice(img_file, fobj=fobj)
      blkdev.create(geo)
    return blkdev


# --- mini test ---
if __name__ == '__main__':
  import sys
  import StringIO
  bdf = BlkDevFactory()
  for a in sys.argv[1:]:
    # open by file
    blkdev = bdf.open(a)
    print(a, blkdev.__class__.__name__)
    blkdev.close()
    # open via fobj
    fobj = file(a,"rb")
    data = fobj.read()
    nobj = StringIO.StringIO(data)
    blkdev = bdf.open("bluna"+a, fobj=nobj)
    print(a, blkdev.__class__.__name__)
    blkdev.close()



import random


class GameDatabaseClient(object):
    def __init__(self, database):
        self.database = database

    def get_or_create_game_id(self, game_uuid):
        cursor = self.database.cursor()
        cursor.execute(
            self.database.query("SELECT id FROM game WHERE uuid = %s"),
            (game_uuid,),
        )
        row = cursor.fetchone()
        if row is None:
            # cursor.execute("INSERT INTO game (uuid, updated) "
            #         "VALUES (%s, utc_timestamp())",
            #         (game_uuid,))
            cursor.execute(
                self.database.query("INSERT INTO game (uuid) VALUES (%s)"),
                (game_uuid,),
            )
            cursor.execute(
                self.database.query("SELECT id FROM game WHERE uuid = %s"),
                (game_uuid,),
            )
            row = cursor.fetchone()
        return row[0]

    def get_game_id(self, game_uuid):
        cursor = self.database.cursor()
        cursor.execute(
            self.database.query("SELECT id FROM game WHERE uuid = %s"),
            (game_uuid,),
        )
        row = cursor.fetchone()
        if row is None:
            raise Exception("game {0} not found".format(game_uuid))
        return row[0]

    def get_ancestor_game_id(self, game_uuid):
        cursor = self.database.cursor()
        cursor.execute(
            self.database.query("SELECT id, parent FROM game WHERE uuid = %s"),
            (game_uuid,),
        )
        row = cursor.fetchone()
        if row is None:
            raise Exception("game {0} not found".format(game_uuid))
        return row[1] or row[0]

    def get_or_create_user_id(self, username):
        cursor = self.database.cursor()
        cursor.execute(
            self.database.query("SELECT id FROM user WHERE username = %s"),
            (username,),
        )
        row = cursor.fetchone()
        if row is None:
            cursor.execute(
                self.database.query("INSERT INTO user (username) VALUES (%s)"),
                (username,),
            )
            cursor.execute(
                self.database.query("SELECT id FROM user WHERE username = %s"),
                (username,),
            )
            row = cursor.fetchone()
        return row[0]

    def insert_game_value(
        self,
        game_id,
        key,
        value,
        status,
        user=1,
        submitted=None,
        update_games=True,
        value_id=None,
        reviewer=None,
    ):
        key = key.strip().lower().replace("-", "_")
        value = value.strip()

        cursor = self.database.cursor()
        # print((game_id, key, value))
        cursor.execute(
            self.database.query(
                "SELECT value FROM value WHERE game = %s AND name = %s AND "
                "status = 1"
            ),
            (game_id, key),
        )
        row = cursor.fetchone()
        # if row is not None and row[0].decode("UTF-8") == value:
        if row is not None and row[0] == value:
            # key - value already active
            if value_id:
                # we must insert this anyway, since a specific value id is
                # wanted in the database
                pass
            else:
                return
        if status == 1:
            cursor.execute(
                self.database.query(
                    "UPDATE value SET status = 0 WHERE game = %s AND name = %s "
                    "AND status = %s"
                ),
                (game_id, key, status),
            )
        if submitted and value_id:
            cursor.execute(
                self.database.query(
                    "INSERT INTO value (id, game, name, value, submitted, "
                    "submitter, status) VALUES (%s, %s, %s, %s, %s, %s, %s)"
                ),
                (value_id, game_id, key, value, submitted, user, status),
            )
            cursor.execute(self.database.query("SELECT max(id) from value"))
        elif reviewer is not None:
            cursor.execute(
                self.database.query(
                    "INSERT INTO value (game, name, value, submitted, "
                    "submitter, status, reviewer, reviewed) VALUES (%s, %s, "
                    "%s, %s, %s, %s, %s, utc_timestamp())"
                ),
                (game_id, key, value, submitted, user, status, reviewer),
            )
        else:
            cursor.execute(
                self.database.query(
                    "INSERT INTO value (game, name, value, submitted, "
                    "submitter, status) VALUES (%s, %s, %s, utc_timestamp(), "
                    "%s, %s)"
                ),
                (game_id, key, value, user, status),
            )

        if (
            update_games
            and status == 1
            and key
            in [
                "game_name",
                "sort_key",
                "variant_name",
                "x_name",
                "parent_uuid",
                "file_list",
                "_status",
                "__source",
                "_downloadable",
                "year",
                "publisher",
                "front_sha1",
                "title_sha1",
                "screen1_sha1",
                "screen2_sha1",
                "screen3_sha1",
                "screen4_sha1",
                "screen5_sha1",
            ]
        ):
            self.update_game(game_id)
        else:
            self.update_game_stamp(game_id)

    def get_final_game_values(self, game_id, recursive=True):
        cursor = self.database.cursor()
        values = {}
        original_game_id = game_id
        while game_id:
            cursor.execute(
                self.database.query(
                    "SELECT name, value FROM value WHERE game = %s AND "
                    "status = 1"
                ),
                (game_id,),
            )
            parent_uuid = ""
            for row in cursor:
                name, value = row
                if not value:
                    continue
                if name[0] == "_":
                    # non-inheritable variables
                    if game_id != original_game_id:
                        continue
                if name == "parent_uuid":
                    parent_uuid = value
                if name == "file_list" and game_id != original_game_id:
                    continue
                if name not in values:
                    values[name] = value
            if not recursive:
                return values
            if parent_uuid:
                print("cannot find parent for", game_id, parent_uuid)
                cursor.execute(
                    self.database.query("SELECT id FROM game WHERE uuid = %s"),
                    (parent_uuid,),
                )
                game_id = cursor.fetchone()[0]
            else:
                game_id = None
        return values

    def get_game_values(self, game_id, status=1):
        return self.get_final_game_values(game_id, False)

    def get_pending_game_values(self, game_id):
        cursor = self.database.cursor()
        values = []
        cursor.execute(
            self.database.query(
                "SELECT name, value FROM value WHERE game = %s and status = -1"
            ),
            (game_id,),
        )
        for row in cursor:
            name, value = row
            values.append((name, value))
        return values

    def get_game_value_data(self, game_id):
        cursor = self.database.cursor()
        values = []
        cursor.execute(
            self.database.query(
                "SELECT value.id, status, name, value, username FROM value, user "
                "WHERE value.submitter = user.id AND game = %s AND "
                "(status = 1 OR status = -1) ORDER BY status DESC, name"
            ),
            (game_id,),
        )
        for row in cursor:
            id, status, name, value, username = row
            values.append((id, status, name, value, username))
        return values

    def update_game(self, game_id):
        cursor = self.database.cursor()
        values = self.get_final_game_values(game_id)
        platform = values.get("platform", "")
        game = values.get("game_name", "")
        sort_key = values.get("sort_key", "")
        variant = values.get("variant_name", "")
        x_name = values.get("x_name", "")
        try:
            status = int(values.get("_status", "0"))
        except:
            status = 0
        parent_uuid = values.get("parent_uuid", "")

        downloadable = values.get("auto_download", "")
        if downloadable:
            # downloadable = "file:" + downloadable
            pass
        if not downloadable:
            downloadable = values.get("manual_download", "")
            # if downloadable:
            #     downloadable = "page:" + downloadable
        if not downloadable:
            downloadable = values.get("__source", "")
            if downloadable.startswith("http"):
                # downloadable = "page:" + downloadable
                pass
            else:
                downloadable = ""

        parent = None
        if parent_uuid:
            cursor.execute(
                self.database.query("SELECT id FROM game WHERE uuid = %s"),
                (parent_uuid,),
            )
            row = cursor.fetchone()
            if row is not None:
                parent = row[0]

        files = 1 if values.get("file_list", "") else 0
        if game and variant:
            name = "{0} ({1})".format(game, variant)
        else:
            name = x_name or game or variant

        if not sort_key:
            sort_key = game
        if sort_key and variant:
            sort_key = "{0} ({1})".format(sort_key, variant)
        else:
            sort_key = x_name or sort_key or variant
        sort_key = sort_key.lower()

        search = name.lower()
        cursor.execute(
            self.database.query(
                "UPDATE game SET game = %s, "
                "variant = %s, name = %s, search = %s, platform = %s, "
                "files = %s, parent = %s, sort_key = %s, status = %s, "
                "downloadable = %s, year = %s, "
                "publisher = %s, front_sha1 = %s, title_sha1 = %s, "
                "screen1_sha1 = %s, screen2_sha1 = %s, screen3_sha1 = %s, "
                "screen4_sha1 = %s, screen5_sha1 = %s, "
                "update_stamp = %s WHERE id = %s"
            ),
            (
                game,
                variant,
                name,
                search,
                platform,
                files,
                parent,
                sort_key,
                status,
                downloadable,
                values.get("year", ""),
                values.get("publisher", ""),
                values.get("front_sha1", ""),
                values.get("title_sha1", ""),
                values.get("screen1_sha1", ""),
                values.get("screen2_sha1", ""),
                values.get("screen3_sha1", ""),
                values.get("screen4_sha1", ""),
                values.get("screen5_sha1", ""),
                create_update_stamp(),
                game_id,
            ),
        )

        cursor.execute(
            self.database.query("SELECT id FROM game WHERE parent = %s"),
            (game_id,),
        )
        game_ids = []
        for row in cursor:
            game_ids.append(row[0])
        for gid in game_ids:
            self.update_game(gid)

    def update_game_stamp(self, game_id):
        cursor = self.database.cursor()
        cursor.execute(
            self.database.query(
                "UPDATE game SET update_stamp = %s WHERE id = %s"
            ),
            (create_update_stamp(), game_id),
        )

    def get_changes(self, start, limit):
        cursor = self.database.cursor()
        cursor.execute(
            self.database.query(
                "SELECT value.id, value.name, value, user.username, submitted, "
                "game.uuid FROM value, user, game WHERE "
                "value.submitter = user.id AND value.game = game.id AND "
                "status = 1 AND value.id >= %s ORDER BY value.id LIMIT %s"
            ),
            (start, limit),
        )
        for row in cursor:
            yield {
                "id": row[0],
                "key": row[1],
                # "value": row[2].decode("UTF-8"),
                "value": row[2],
                "submitter": row[3],
                "submitted": str(row[4]),
                "game": row[5],
            }

    def get_ratings(self, start, limit, user_id=None):
        cursor = self.database.cursor()
        if user_id is not None:
            cursor.execute(
                self.database.query(
                    "SELECT game.uuid, game_rating.work_rating, "
                    "game_rating.like_rating, game_rating.updated FROM game, "
                    "game_rating WHERE game_rating.game = game.id AND "
                    "game_rating.updated >= %s AND game_rating.user = %s "
                    "LIMIT %s"
                ),
                (start, user_id, limit),
            )
        else:
            cursor.execute(
                self.database.query(
                    "SELECT uuid, work_rating, like_rating, updated FROM game "
                    "WHERE updated >= %s LIMIT %s"
                ),
                (start, limit),
            )
        for row in cursor:
            yield {
                "game": row[0],
                "work": row[1],
                "like": row[2],
                "updated": str(row[3]),
            }

    def get_user_id(self, username, password):
        cursor = self.database.cursor()
        cursor.execute(
            self.database.query(
                "SELECT id FROM user WHERE username = %s and password = %s"
            ),
            (username, password),
        )
        row = cursor.fetchone()
        if row is None:
            return None
        return row[0]

    def get_last_change_id(self):
        cursor = self.database.cursor()
        result = cursor.execute("SELECT max(id) from value")
        last_id = result.fetchone()[0]
        if not last_id:
            return 0
        return last_id

    def get_game_database_version(self):
        cursor = self.database.cursor()
        result = cursor.execute(
            self.database.query("SELECT games_version from metadata")
        )
        return result.fetchone()[0]

    def set_game_database_version(self, version):
        cursor = self.database.cursor()
        if self.get_game_database_version() == version:
            # already correct version
            return
        cursor.execute(
            self.database.query("UPDATE metadata SET games_version = %s"),
            (version,),
        )

    def delete_old_values(self):
        cursor = self.database.cursor()
        cursor.execute(
            self.database.query("DELETE FROM value WHERE status = 0")
        )

    def clear_database(self):
        cursor = self.database.cursor()
        cursor.execute(self.database.query("DELETE FROM value"))
        cursor.execute(self.database.query("DELETE FROM game"))
        cursor.execute(self.database.query("DELETE FROM user"))


def create_update_stamp():
    return random.randint(0, 2 ** 32 - 1)



import os

from fsgs.drivers.dolphindriver import DolphinDriver, DolphinInputMapper
from fsgs.drivers.gamedriver import GameDriver, Emulator
from fsgs.platform import Platform
from fsgs.platforms.loader import SimpleLoader

NDS_PLATFORM_NAME = "Nintendo DS"
NDS_CONTROLLER = {
    "type": "gamepad",
    "description": "Gamepad",
    "mapping_name": "ngc",
}


class NintendoDSPlatform(Platform):
    PLATFORM_NAME = NDS_PLATFORM_NAME

    def driver(self, fsgc):
        return NintendoDSDriver(fsgc)

    def loader(self, fsgc):
        return NintendoDSLoader(fsgc)


class NintendoDSLoader(SimpleLoader):
    pass


class NintendoDSDriver(GameDriver):
    PORTS = [{"description": "Controller 1", "types": [NDS_CONTROLLER]}]

    def __init__(self, fsgc):
        super().__init__(fsgc)
        self.helper = NintendoDSHelper(self.options)
        self.emulator = Emulator("desmume")
        self.emulator.allow_system_emulator = True

    def prepare(self):
        rom_path = self.get_game_file()
        self.emulator.args.append(rom_path)

    def finish(self):
        pass


class NintendoDSHelper:
    def __init__(self, options):
        self.options = options



import traceback
import weakref

import fsui
from fsbc.util import unused
from launcher.launcher_config import LauncherConfig
from launcher.launcher_settings import LauncherSettings
from launcher.ui.config.expand import expand_config, AbstractExpandFunctions
from launcher.ui.config.model import ImplicitConfig, normalize


class ImplicitConfigHandler:
    def __init__(self, parent):
        self.parent = weakref.ref(parent)
        LauncherConfig.add_listener(self)
        LauncherSettings.add_listener(self)
        self.dirty = True
        self.do_update()
        parent.destroyed.connect(self.on_parent_destroyed)

    def on_parent_destroyed(self):
        print("ImplicitConfigHandler.on_parent_destroyed")
        LauncherConfig.remove_listener(self)
        LauncherSettings.remove_listener(self)

    def update(self):
        self.dirty = True
        fsui.call_after(self.do_update)

    def do_update(self):
        if not self.dirty:
            return
        self.dirty = False
        # print("ImplicitConfigHandler.do_update")
        implicit = ImplicitConfig(ConfigProxy(), SettingsProxy())
        # failed = False
        try:
            expand_config(implicit, ExpandFunctions())
        except Exception:
            traceback.print_exc()
            print("expand_config failed")
            # failed = True
        implicit_config = {
            key: ""
            for key in LauncherConfig.keys()
            if key.startswith("__implicit_")
        }
        for key, value in implicit.items():
            implicit_config["__implicit_" + key] = value
        LauncherConfig.set_multiple(list(implicit_config.items()))

        if self.parent().config_browser:
            self.parent().config_browser.update_from_implicit(implicit)

    def on_config(self, key, value):
        if key.startswith("__implicit_"):
            # The implicit key/values are already set by us.
            return
        unused(value)
        self.update()

    def on_setting(self, key, value):
        unused(key)
        unused(value)
        self.update()


class SettingsProxy:
    @staticmethod
    def get(key):
        return LauncherSettings.get(key)

    def __getitem__(self, item):
        return self.get(item)

    def __setitem__(self, key, value):
        LauncherSettings.set(key, value)


class ConfigProxy:
    @staticmethod
    def get(key):
        return LauncherConfig.get(key)

    def __getitem__(self, item):
        return self.get(item)

    def __setitem__(self, key, value):
        LauncherConfig.set(key, value)


class ExpandFunctions(AbstractExpandFunctions):
    @staticmethod
    def matches(a, b):
        a = normalize(a)
        if isinstance(b, list):
            for b_item in b:
                if a == normalize(b_item):
                    return True
            return False
        return a == normalize(b)

    @staticmethod
    def fail(message):
        pass

    @staticmethod
    def warning(message):
        # warnings.append(message)
        pass

    @staticmethod
    def lower(s):
        return s.lower()



import logging
import os
import warnings
from configparser import ConfigParser
from functools import lru_cache
from typing import Dict, List, Optional

import fsboot
from fsbc.paths import Paths
from fsbc.settings import Settings
from fsbc.system import windows, macosx
from fsbc.user import get_common_documents_dir
from fsbc.user import get_documents_dir
from fsgs import openretro


class FSGSDirectories(object):
    @classmethod
    def initialize(cls):
        warnings.warn(DeprecationWarning)
        logging.debug("[DEPRECATED] FSGSDirectories.initialize")
        cls._initialize()

    @classmethod
    def _initialize(cls):
        if hasattr(cls, "_initialized") and cls._initialized:
            return
        logging.debug("FSGSDirectories._initialize")
        cls.get_base_dir()
        cls._initialized = True

    @classmethod
    @lru_cache()
    def get_base_dir(cls) -> str:
        path = fsboot.base_dir()
        # Configuration and file database depends on path normalization,
        # especially for cross-platform portable mode.
        path = Paths.get_real_case(path)
        print("FSGSDirectories.get_base_dir =", repr(path))
        return path

    @classmethod
    @lru_cache()
    def portable_config(cls):
        portable_ini = os.path.join(cls.get_base_dir(), "Portable.ini")
        if os.path.exists(portable_ini):
            cp = ConfigParser()
            cp.read(portable_ini, encoding="UTF-8")
            if cp.has_section("directories"):
                return {k: v for (k, v) in cp.items("directories")}
        return {}

    @classmethod
    @lru_cache()
    def portable_dir(cls, name):
        config = cls.portable_config()
        try:
            path = config[name]
        except KeyError:
            return None
        return os.path.join(cls.get_base_dir(), path)

    @classmethod
    @lru_cache()
    def get_configurations_dir(cls):
        path = cls.portable_dir("configurations_dir")
        if not path:
            if openretro:
                path = os.path.join(cls.get_data_dir(), "Configs")
            else:
                path = os.path.join(cls.get_base_dir(), "Configurations")
        if not os.path.exists(path):
            os.makedirs(path)
        path = Paths.get_real_case(path)
        return path

    @classmethod
    @lru_cache()
    def get_controllers_dir(cls):
        path = cls.portable_dir("controllers_dir")
        if not path:
            if openretro:
                path = os.path.join(cls.get_data_dir(), "Devs", "Joysticks")
            else:
                path = os.path.join(cls.get_base_dir(), "Controllers")
        if not os.path.exists(path):
            os.makedirs(path)
        path = Paths.get_real_case(path)
        return path

    @classmethod
    @lru_cache()
    def get_kickstarts_dir(cls):
        path = cls.portable_dir("kickstarts_dir")
        if not path:
            if openretro:
                path = os.path.join(cls.media_dir(), "ROMs")
            else:
                path = os.path.join(cls.get_base_dir(), "Kickstarts")
        if not os.path.exists(path):
            os.makedirs(path)
        path = Paths.get_real_case(path)
        return path

    @classmethod
    @lru_cache()
    def downloads_dir(cls):
        path = cls.portable_dir("downloads_dir")
        if not path:
            path = os.path.join(cls.get_base_dir(), "Downloads")
        if not os.path.exists(path):
            os.makedirs(path)
        path = Paths.get_real_case(path)
        return path

    @classmethod
    def ensure_downloads_dir(cls):
        path = cls.downloads_dir()
        if not os.path.exists(path):
            os.makedirs(path)
        return path

    @classmethod
    @lru_cache()
    def get_floppies_dir(cls):
        path = cls.portable_dir("floppies_dir")
        if not path:
            if openretro:
                path = os.path.join(cls.media_dir(), "Floppies")
            else:
                path = os.path.join(cls.get_base_dir(), "Floppies")
        if not os.path.exists(path):
            os.makedirs(path)
        path = Paths.get_real_case(path)
        return path

    @classmethod
    @lru_cache()
    def get_hard_drives_dir(cls):
        path = cls.portable_dir("hard_drives_dir")
        if not path:
            if openretro:
                path = os.path.join(cls.media_dir(), "HardDrives")
            else:
                path = os.path.join(cls.get_base_dir(), "Hard Drives")
        if not os.path.exists(path):
            os.makedirs(path)
        path = Paths.get_real_case(path)
        return path

    @classmethod
    @lru_cache()
    def get_cdroms_dir(cls):
        path = cls.portable_dir("cdroms_dir")
        if not path:
            if openretro:
                path = os.path.join(cls.media_dir(), "CD-ROMs")
            else:
                path = os.path.join(cls.get_base_dir(), "CD-ROMs")
        if not os.path.exists(path):
            os.makedirs(path)
        path = Paths.get_real_case(path)
        return path

    @classmethod
    @lru_cache()
    def get_logs_dir(cls):
        path = cls.portable_dir("logs_dir")
        if not path:
            path = os.path.join(cls.get_cache_dir(), "Logs")
        if not os.path.exists(path):
            os.makedirs(path)
        path = Paths.get_real_case(path)
        return path

    @classmethod
    @lru_cache()
    def get_data_dir(cls):
        path = cls.portable_dir("data_dir")
        if not path:
            path = os.path.join(cls.get_base_dir(), "Data")
        if not os.path.exists(path):
            os.makedirs(path)
        path = Paths.get_real_case(path)
        return path

    @classmethod
    @lru_cache()
    def databases_dir(cls):
        path = os.path.join(cls.get_data_dir(), "Databases")
        if not os.path.exists(path):
            os.makedirs(path)
        return path

    @classmethod
    @lru_cache()
    def images_dir(cls):
        path = os.path.join(cls.get_data_dir(), "Images")
        if not os.path.exists(path):
            os.makedirs(path)
        return path

    @classmethod
    @lru_cache()
    def media_dir(cls):
        path = cls.portable_dir("media_dir")
        if not path:
            path = os.path.join(cls.get_base_dir(), "Media")
        if not os.path.exists(path):
            os.makedirs(path)
        path = Paths.get_real_case(path)
        return path

    @classmethod
    @lru_cache()
    def images_dir_for_sha1(cls, sha1):
        path = os.path.join(cls.images_dir(), sha1[:2])
        # if not os.path.exists(path):
        #     os.makedirs(path)
        return path

    @classmethod
    @lru_cache()
    def get_launcher_dir(cls):
        return cls.get_data_dir()

    @classmethod
    @lru_cache()
    def get_titles_dir(cls):
        path = cls.portable_dir("titles_dir")
        if not path:
            if openretro:
                path = os.path.join(cls.get_data_dir(), "Titles")
            else:
                path = os.path.join(cls.get_base_dir(), "Titles")
        return path

    @classmethod
    @lru_cache()
    def saves_dir(cls):
        path = cls.portable_dir("saves_dir")
        if not path:
            path = os.path.join(cls.get_data_dir(), "Saves")
        if not os.path.exists(path):
            os.makedirs(path)
        path = Paths.get_real_case(path)
        return path

    @classmethod
    @lru_cache()
    def get_save_states_dir(cls):
        path = cls.portable_dir("save_states_dir")
        if not path:
            if openretro:
                path = os.path.join(cls.get_data_dir(), "Saves")
            else:
                path = os.path.join(cls.get_base_dir(), "Save States")
        if not os.path.exists(path):
            os.makedirs(path)
        path = Paths.get_real_case(path)
        return path

    @classmethod
    @lru_cache()
    def screenshots_output_dir(cls):
        path = Settings.instance()["screenshots_output_dir"]
        if not path:
            path = cls.portable_dir("screenshots_output_dir")
        if not path:
            path = os.path.join(get_documents_dir(), "Screenshots")
        if not os.path.exists(path):
            os.makedirs(path)
        return path

    @classmethod
    @lru_cache()
    def get_screenshots_dir(cls):
        path = cls.portable_dir("screenshots_dir")
        if not path:
            if openretro:
                path = os.path.join(cls.get_data_dir(), "Screenshots")
            else:
                path = os.path.join(cls.get_base_dir(), "Screenshots")
        return path

    @classmethod
    @lru_cache()
    def get_images_dir(cls):
        cls._initialize()
        path = cls.get_base_dir()
        return path

    @classmethod
    @lru_cache()
    def get_covers_dir(cls):
        path = cls.portable_dir("covers_dir")
        if not path:
            if openretro:
                path = os.path.join(cls.get_data_dir(), "Covers")
            else:
                path = os.path.join(cls.get_base_dir(), "Covers")
        return path

    @classmethod
    @lru_cache()
    def get_themes_dir(cls):
        path = cls.portable_dir("themes_dir")
        if not path:
            if openretro:
                path = os.path.join(cls.get_data_dir(), "Themes")
            else:
                path = os.path.join(cls.get_base_dir(), "Themes")
        if not os.path.exists(path):
            os.makedirs(path)
        return path

    @classmethod
    def get_plugins_dir(cls):
        return cls.get_system_dir()

    @classmethod
    @lru_cache()
    def get_system_dir(cls):
        path = os.path.join(cls.get_base_dir(), "System")
        if not os.path.exists(path):
            os.makedirs(path)
        return path

    @classmethod
    @lru_cache()
    def get_cache_dir(cls):
        path = cls.portable_dir("cache_dir")
        if not path:
            path = os.path.join(cls.get_base_dir(), "Cache")
        if not os.path.exists(path):
            os.makedirs(path)
        if not os.path.exists(os.path.join(path, ".nobackup")):
            open(os.path.join(path, ".nobackup"), "w").close()
        return path

    @classmethod
    @lru_cache()
    def get_files_dirs(cls):
        paths = [cls.get_floppies_dir(), cls.get_cdroms_dir()]
        return paths

    @classmethod
    @lru_cache()
    def get_titles_dirs(cls):
        paths = [cls.get_titles_dir()]
        return paths

    @classmethod
    @lru_cache()
    def get_screenshots_dirs(cls):
        paths = [cls.get_screenshots_dir()]
        return paths

    @classmethod
    @lru_cache()
    def get_images_dirs(cls):
        paths = [cls.get_images_dir()]
        return paths

    @classmethod
    @lru_cache()
    def get_covers_dirs(cls):
        paths = [cls.get_covers_dir()]
        return paths

    @classmethod
    @lru_cache()
    def get_themes_dirs(cls):
        paths = [cls.get_themes_dir()]
        return paths

    @classmethod
    def get_amiga_forever_directories(cls):
        paths = []
        if fsboot.is_portable():
            # Portable version, don't try to find ROM files outside the
            # portable directory by default.
            pass
        else:
            if windows:
                path = get_common_documents_dir()
                path = os.path.join(path, "Amiga Files")
                if os.path.exists(path):
                    paths.append(path)
            else:
                path = os.path.expanduser(
                    "~/.wine/drive_c/users/Public/Documents/Amiga Files"
                )
                if os.path.exists(path):
                    paths.append(path)
        path = cls.get_base_dir()
        path = os.path.join(path, "AmigaForever", "Amiga Files")
        if os.path.exists(path):
            paths.append(path)
        return paths

    @classmethod
    def get_default_search_path(cls):
        paths = []
        path = cls.get_base_dir()
        paths.append(path)
        for path in cls.get_amiga_forever_directories():
            paths.append(path)
        return paths



import sys
from fsgs.plugins.pluginmanager import PluginManager

"""
MAME-FS launcher script used for testing.
"""


def app_main():
    executable = PluginManager.instance().find_executable("mame-fs")
    process = executable.popen(sys.argv[1:])
    process.wait()



import fsui
from launcher.ui.behaviors.platformbehavior import AmigaEnableBehavior
from ...launcher_config import LauncherConfig
from ...i18n import gettext
from ..HelpButton import HelpButton


class WHDLoadGroup(fsui.Panel):
    def __init__(self, parent):
        fsui.Panel.__init__(self, parent)
        AmigaEnableBehavior(self)
        self.layout = fsui.VerticalLayout()
        self.text_field = fsui.TextField(self, "")
        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True, margin=10)
        hori_layout.add(self.text_field, expand=True)
        self.help_button = HelpButton(
            self, "https://fs-uae.net/whdload-support"
        )
        hori_layout.add(self.help_button, margin_left=10)
        self.initialize_from_config()
        self.set_config_handlers()

    def initialize_from_config(self):
        self.on_config("x_whdload_args", LauncherConfig.get("x_whdload_args"))

    def set_config_handlers(self):
        self.text_field.on_changed = self.on_text_changed
        LauncherConfig.add_listener(self)

    def on_destroy(self):
        print("on_destroy")
        LauncherConfig.remove_listener(self)

    def on_config(self, key, value):
        if key == "x_whdload_args":
            if value != self.text_field.get_text():
                self.text_field.set_text(value)

    def on_text_changed(self):
        LauncherConfig.set("x_whdload_args", self.text_field.get_text())



import os
import time
import threading
from fsbc.system import windows, macosx


class Joystick(object):
    @classmethod
    def _get_joystick_hash(cls):

        if windows:
            # FIXME: WARNING: DISABLING FOR NOW, SINCE STARTING THIS CONSOLE
            # APPLICATION CAUSES NOTICEABLE DELAYS (AND AN HOURGLASS)
            # - Try to spawn a "server" communicating with us via STDOUT, and
            # let the server periodically start a joystick polling app.
            return hash("FIXME")

            # # using an external app here, because windows seems to cache
            # # device information while the program is running...
            #
            # path = os.path.join(pyapp.plugins.get_plugin(
            #         'no.fengestad.input.platform').get_bin_dir(),
            #         'dinput8.exe')
            # # Must specify all files (stdin/stdout/stderr) when running in
            # # GUI mode on Windows
            #
            # si = subprocess.STARTUPINFO()
            # si.dwFlags = subprocess.STARTF_USESHOWWINDOW
            # si.wShowWindow = subprocess.SW_HIDE
            #
            # p = subprocess.Popen([path], stdin=subprocess.PIPE,
            #         stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
            #         startupinfo=si)
            # p.stdin.close()
            # # lines = 0
            # # for line in p.stdout.readlines():
            # #     if line[0] != '\n' and line[0] != '\r':
            # #         lines += 1
            # # # dinput8.exe prints two lines per device
            # # return lines // 2
            # data = p.stdout.read()
            # # print(data)
            # return hash(data)

        elif macosx:
            # FIXME: Implement
            # return hash(0)
            return hash("FIXME")

        else:
            if os.path.exists("/dev/input"):
                return hash(tuple(os.listdir("/dev/input")))
            return hash("FIXME")

    _hash = None

    @classmethod
    def get_joystick_hash(cls):
        # while cls._hash is None:
        #     time.sleep(0.01)
        if cls._hash is None:
            cls._hash = cls._get_joystick_hash()
            # print("FIXME: not starting thread in get_joystick_hash")
            t = threading.Thread(target=cls.thread, name="JoystickThread")
            t.daemon = True
            t.start()
        return cls._hash

    @classmethod
    def thread(cls):
        try:
            cls._thread()
        except Exception:
            import traceback

            traceback.print_exc()

    @classmethod
    def _thread(cls):
        while True:
            time.sleep(1.0)
            cls._hash = cls._get_joystick_hash()



from fsui.qt import Qt
from .FileDialog import FileDialog


class DirDialog(FileDialog):
    def __init__(self, parent=None, message="", directory=""):
        FileDialog.__init__(self, parent, message, directory, dir_mode=True)
        self.setAttribute(Qt.WA_DeleteOnClose)



import time
from functools import lru_cache

import fsbc.settings
from fsgs.filedatabase import FileDatabase
from fsgs.LockerDatabase import LockerDatabase
from fsgs.download import Downloader
from fsgs.network import openretro_url_prefix
from fsgs.ogd.base import SynchronizerBase
from fsgs.ogd.context import SynchronizerContext
from fsgs.res import gettext


def is_locker_enabled():
    return fsbc.settings.get("database_locker") == "1"


def open_locker_uri(uri):
    sha1 = uri[9:]
    assert len(sha1) == 40
    context = SynchronizerContext()
    url = "{0}/api/locker/{1}".format(openretro_url_prefix(), sha1)
    path = Downloader.cache_file_from_url(
        url, auth=(context.username, context.password)
    )
    return path


class LockerSynchronizer(SynchronizerBase):
    def __init__(self, *args, **kwargs):
        SynchronizerBase.__init__(self, *args, **kwargs)

    def synchronize(self):
        if not is_locker_enabled():
            return
        if "locker" not in self.context.meta:
            # We haven't looked up synchronization information from the server,
            # that probably means we didn't want to synchronize with the
            # server now, therefore we just return.
            return
        if self.stop_check():
            return
        database = LockerDatabase.instance()
        sync_version = database.get_sync_version()
        if sync_version == self.context.meta["locker"]["sync"]:
            print("[SYNC] Locker data already up to date")
            return

        self.set_status(gettext("Fetching locker data..."))
        data = self.fetch_data("/api/locker-sync/1")
        assert len(data) % 20 == 0
        self.set_status(gettext("Updating locker data..."))
        database.clear()
        k = 0
        while k < len(data):
            sha1_bytes = data[k : k + 20]
            database.add_sha1_binary(sha1_bytes)
            k += 20

        database.set_sync_version(self.context.meta["locker"]["sync"])
        self.set_status(gettext("Committing locker data..."))
        self.update_file_database_timestamps()
        database.commit()

    def update_file_database_timestamps(self):
        # This isn't very elegant, but in order to force the game scanner to
        # refresh the game list based on files, we update the stamps in the
        # file database. Also, since we haven't keep track of additions /
        # deletions, we set both stamps for now...
        file_database = FileDatabase.instance()
        file_database.last_file_insert = time.time()
        file_database.last_file_delete = time.time()
        file_database.update_last_event_stamps()
        file_database.commit()



from fsui import Panel, VerticalLayout
from launcher.ui.behaviors.platformbehavior import AMIGA_PLATFORMS
from launcher.ui.skin import Skin
from launcher.ui.options import ConfigWidgetFactory


class ConfigPanel(Panel):
    def __init__(self, parent):
        super().__init__(parent)
        Skin.set_background_color(self)
        self.layout = VerticalLayout()
        self.config_widget_factory = ConfigWidgetFactory()

    def add_option(self, name, fill=True):
        self.layout.add(
            self.config_widget_factory.create(self, name),
            fill=fill,
            margin=10,
            margin_bottom=0,
        )

    def add_amiga_option(self, name, fill=True, parent=None):
        if parent is None:
            parent = self
        parent.layout.add(
            self.config_widget_factory.create(
                parent, name, platforms=AMIGA_PLATFORMS
            ),
            fill=fill,
            margin=10,
            margin_bottom=0,
        )



from __future__ import absolute_import
from __future__ import print_function

from .Block import Block

class FileDataBlock(Block):
  def __init__(self, blkdev, blk_num):
    Block.__init__(self, blkdev, blk_num, is_type=Block.T_DATA)
  
  def set(self, data):
    self._set_data(data)
    self._read()
  
  def read(self):
    self._read_data()
    self._read()
  
  def _read(self):
    Block.read(self)
    if not self.valid:
      return False
    
    # FileData fields
    self.hdr_key = self._get_long(1)
    self.seq_num = self._get_long(2)
    self.data_size = self._get_long(3)
    self.next_data = self._get_long(4)
    
    self.valid = True
    return self.valid

  def create(self, hdr_key, seq_num, data, next_data):
    Block.create(self)
    self.hdr_key = hdr_key
    self.seq_num = seq_num
    self.data_size = len(data)
    self.next_data = next_data
    self.contents = data
    
  def write(self):
    Block._create_data(self)
    self._put_long(1, self.hdr_key)
    self._put_long(2, self.seq_num)
    self._put_long(3, self.data_size)
    self._put_long(4, self.next_data)
    if self.contents != None:
      self.data[24:24+self.data_size] = self.contents
    Block.write(self)

  def get_block_data(self):
    return self.data[24:24+self.data_size]
  
  def dump(self):
    Block.dump(self,"FileData")
    print(" hdr_key:    %d" % self.hdr_key)
    print(" seq_num:    %d" % self.seq_num)
    print(" data size:  %d" % self.data_size)
    print(" next_data:  %d" % self.next_data)
  



import re

tag = "$VER:"

def find(data):
  off = data.find(tag)
  if off == -1:
    return None
  start = off + len(tag)
  end = start
  size = len(data)
  while end < size:
    if data[end] == chr(0):
      break
    end += 1
  return data[start:end].strip()
  
def get_version(str):
  m = re.search('\s(\d+)\.(\d+)\s', str)
  if m == None:
    return None
  else:
    return (int(m.group(1)), int(m.group(2)))



from fsgs.plugins.pluginexecutablefinder import PluginExecutableFinder
import os
import subprocess

from fsbc import settings
from fsgs.option import Option
from .fsuae import FSUAE


class FSUAEDeviceHelper(object):
    @classmethod
    def start_with_args(cls, args, **kwargs):
        print("FSUAE.start_with_args:", args)
        exe = PluginExecutableFinder().find_executable("fs-uae-device-helper")
        if exe is None:
            raise Exception("Could not find fs-uae-device-helper executable")
        print("current dir (cwd): ", os.getcwd())
        print("using fs-uae executable:", exe)
        args = [exe] + args
        print(args)
        env = os.environ.copy()
        if settings.get(Option.FAKE_JOYSTICKS):
            try:
                fake_joysticks = int(settings.get(Option.FAKE_JOYSTICKS))
            except ValueError:
                print(
                    "WARNING: fake_joysticks contains invalid value",
                    repr(settings.get(Option.FAKE_JOYSTICKS)),
                )
            else:
                env["FSGS_FAKE_JOYSTICKS"] = str(fake_joysticks)
        FSUAE.add_environment_from_settings(env)
        process = subprocess.Popen(
            args,
            env=env,
            stdin=subprocess.PIPE,
            stderr=subprocess.PIPE,
            **kwargs
        )
        return process



#!/usr/bin/env python3

import json
from PIL import Image


# def fix(name, w, h):
#    with open(name + ".raw", "rb") as f:
#        data = f.read()
#    im = Image.fromstring("RGBA", (w, h), data)
#    im.save(name + ".png")
#
#
# fix("title_font", 1024, 132)
# fix("menu_font", 1024, 132)


def white(pixels, x, y):
    # print(pixels[x, y])
    return pixels[x, y][0] > 128


def calculate(name):
    char_x = []
    char_y = []
    char_w = []
    char_h = []
    im = Image.open(name + ".png")
    pixels = im.load()
    x, y = 0, 0
    w, h = im.size
    char = -1
    in_x = False
    in_y = False
    rows = []
    y = 1
    while y < h:
        if not in_y and white(pixels, x, y):
            in_y = True
            rows.append([y])
        elif in_y and not white(pixels, x, y):
            in_y = False
            rows[-1].append(y - rows[-1][0])
        y += 1
    if len(rows[-1]) == 1:
        rows[-1].append(y - rows[-1][0])
    print(rows)
    for row in rows:
        x = 1
        y = row[0] - 1
        while x < w:
            if not in_x and white(pixels, x, y):
                in_x = True
                char += 1
                char_x.append(x)
                char_y.append(y + 1)
                char_h.append(row[1])
            elif in_x and not white(pixels, x, y):
                in_x = False
                char_w.append(x - char_x[-1])
            x += 1
    with open(name + ".json", "w") as f:
        json.dump({"x": char_x, "y": char_y, "w": char_w, "h": char_h}, f)


calculate("title_font")
calculate("menu_font")



from __future__ import absolute_import
from __future__ import print_function

import struct
import ctypes
from ..TimeStamp import TimeStamp

class Block:
  # mark end of block list
  no_blk = 0xffffffff
  
  # special blocks
  RDSK = 0x5244534b # Rigid Disk Block
  BADB = 0x42414442 # Bad Blocks Block
  PART = 0x50415254 # Partition Block
  FSHD = 0x46534844 # FileSystem Header Block
  LSEG = 0x4c534547 # LoadSeg Block
  
  # block types
  T_SHORT = 2
  T_DATA = 8
  T_LIST = 16
  T_DIR_CACHE = 33
  T_COMMENT = 64
  # block sub types
  ST_ROOT = 1
  ST_USERDIR = 2
  ST_FILE = -3 & 0xffffffff
  
  def __init__(self, blkdev, blk_num, is_type=0, is_sub_type=0, chk_loc=5):
    self.valid = False
    self.blkdev = blkdev
    self.blk_num = blk_num
    self.block_longs = blkdev.block_longs
    self.type = 0
    self.sub_type = 0
    self.data = None
    self.is_type = is_type
    self.is_sub_type = is_sub_type
    self.chk_loc = chk_loc
  
  def __str__(self):
    return "%s:@%d" % (self.__class__.__name__, self.blk_num)
  
  def create(self):
    self.type = self.is_type
    self.sub_type = self.is_sub_type
  
  def is_root_block(self):
    return self.type == Block.T_SHORT and self.sub_type == Block.ST_ROOT
    
  def is_user_dir_block(self):
    return self.type == Block.T_SHORT and self.sub_type == Block.ST_USERDIR
  
  def is_file_header_block(self):
    return self.type == Block.T_SHORT and self.sub_type == Block.ST_FILE
  
  def is_file_list_block(self):
    return self.type == Block.T_LIST and self.sub_type == Block.ST_FILE
  
  def is_file_data_block(self):
    return self.type == Block.T_DATA
  
  def is_comment_block(self):
    return self.type == Block.T_COMMENT
  
  def read(self):
    if self.data == None:
      self._read_data()
    self._get_types()
    self._get_chksum()
    self.valid = self.valid_types and self.valid_chksum
    
  def write(self):
    if self.data == None:
      self._create_data()
    self._put_types()
    self._put_chksum()
    self._write_data()
  
  def _set_data(self, data):
    self.data = data
  
  def _read_data(self):
    data = self.blkdev.read_block(self.blk_num)
    if len(data) != self.blkdev.block_bytes:
      raise ValueError("Invalid Block Data: size=%d but expected %d" % (len(data), self.blkdev.block_bytes))
    self._create_data()
    self.data[:] = data
  
  def _write_data(self):
    if self.data != None:
      self.blkdev.write_block(self.blk_num, self.data)
  
  def _free_data(self):
    self.data = None
  
  def _create_data(self):
    num_bytes = self.blkdev.block_bytes
    self.data = ctypes.create_string_buffer(num_bytes)
  
  def _put_long(self, num, val):
    if num < 0:
      num = self.block_longs + num
    struct.pack_into(">I",self.data,num*4,val)
  
  def _get_long(self, num):
    if num < 0:
      num = self.block_longs + num
    return struct.unpack_from(">I",self.data,num*4)[0]

  def _put_slong(self, num, val):
    if num < 0:
      num = self.block_longs + num
    struct.pack_into(">i",self.data,num*4,val)
  
  def _get_slong(self, num):
    if num < 0:
      num = self.block_longs + num
    return struct.unpack_from(">i",self.data,num*4)[0]

  def _get_types(self):
    self.type = self._get_long(0)
    self.sub_type = self._get_long(-1)
    self.valid_types = True
    if self.is_type != 0:
      if self.type != self.is_type:
        self.valid_types = False
    if self.is_sub_type != 0:
      if self.sub_type != self.is_sub_type:
        self.valid_types = False
  
  def _put_types(self):
    if self.is_type != 0:
      self._put_long(0, self.is_type)
    if self.is_sub_type != 0:
      self._put_long(-1, self.is_sub_type)
  
  def _get_chksum(self):
    self.got_chksum = self._get_long(self.chk_loc)
    self.calc_chksum = self._calc_chksum()
    self.valid_chksum = self.got_chksum == self.calc_chksum
  
  def _put_chksum(self):
    self.calc_chksum = self._calc_chksum()
    self.got_chksum = self.calc_chksum
    self.valid_chksum = True
    self._put_long(self.chk_loc, self.calc_chksum)
  
  def _calc_chksum(self):
    chksum = 0
    for i in xrange(self.block_longs):
      if i != self.chk_loc:
        chksum += self._get_long(i)
    return (-chksum) & 0xffffffff
  
  def _get_timestamp(self, loc):
    days = self._get_long(loc)
    mins = self._get_long(loc+1)
    ticks = self._get_long(loc+2)
    return TimeStamp(days, mins, ticks)
  
  def _put_timestamp(self, loc, ts):
    if ts == None:
      ts = TimeStamp()
    self._put_long(loc, ts.days)
    self._put_long(loc+1, ts.mins)
    self._put_long(loc+2, ts.ticks)
  
  def _get_bstr(self, loc, max_size):
    if loc < 0:
      loc = self.block_longs + loc
    loc = loc * 4
    size = ord(self.data[loc])
    if size > max_size:
      return None
    if size == 0:
      return ""
    name = self.data[loc+1:loc+1+size]
    return name
  
  def _put_bstr(self, loc, max_size, bstr):
    if bstr == None:
      bstr = ""
    n = len(bstr)
    if n > max_size:
      bstr = bstr[:max_size]
    if loc < 0:
      loc = self.block_longs + loc
    loc = loc * 4
    self.data[loc] = chr(len(bstr))
    if len(bstr) > 0:
      self.data[loc+1:loc+1+len(bstr)] = bstr
  
  def _get_cstr(self, loc, max_size):
    n = 0
    s = ""
    loc = loc * 4
    while n < max_size:
      c = self.data[loc+n]
      if ord(c) == 0:
        break
      s += c
      n += 1
    return s
    
  def _put_cstr(self, loc, max_size, cstr):
    if cstr == None:
      cstr = ""
    n = min(max_size, len(cstr))
    loc = loc * 4
    if n > 0:
      self.data[loc:loc+n] = cstr
  
  def _dump_ptr(self, ptr):
    if ptr == self.no_blk:
      return "none"
    else:
      return "%d" % ptr
  
  def dump(self, name, details=True):
    print("%sBlock(%d):" % (name, self.blk_num))
    if details:
      print(" types:     %x/%x (valid: %x/%x)" % (self.type, self.sub_type, self.is_type, self.is_sub_type))
      print(" chksum:    0x%08x (got) 0x%08x (calc)" % (self.got_chksum, self.calc_chksum))
      print(" valid:     %s" % self.valid)
    


import gzip
import lzma
import io
import os
import traceback
from io import BytesIO

from fspy.zipfile import ZipFile

# This list is also used by the filescanner to add to recognized file
# extensions.
archive_extensions = [".zip", ".rp9"]

try:
    from lhafile import LhaFile
except ImportError:
    traceback.print_exc()
    print("[ARCHIVE] LhaFile module import problem")
    LhaFile = None
else:
    archive_extensions.append(".lha")

try:
    from fsbc.seven_zip_file import SevenZipFile
except ImportError:
    traceback.print_exc()
    print("[ARCHIVE] SevenZipFile module import problem")
    SevenZipFile = None
else:
    archive_extensions.append(".7z")

archive_extensions_gz = [".gz", ".adz", ".roz"]
archive_extensions.extend(archive_extensions_gz)

archive_extensions_xz = [".xz"]
archive_extensions.extend(archive_extensions_xz)


# FIXME: getinfo needs to work for everything (return dummy data if necessary)


class ZipHandler(object):
    def __init__(self, path):
        self.path = path
        self.zip = ZipFile(self.path, "r")

    def exists(self, name):
        try:
            self.zip.getinfo(name)
        except KeyError:
            return False
        else:
            return True

    def getinfo(self, name):
        # Needs tests!!
        return self.zip.getinfo(name)

    def list_files(self, sub_path):
        if sub_path:
            return
        return self.zip.namelist()

    def open(self, name):
        return self.zip.open(name)


class GzipHandler(object):
    def __init__(self, path):
        self.path = path
        name, ext = os.path.splitext(os.path.basename(path))
        ext = ext.lower()
        if ext == ".gz":
            self.name = name
        elif ext == ".adz":
            self.name = name + ".adf"
        elif ext == ".roz":
            self.name = name + ".rom"
        else:
            raise Exception(
                "Unexpected extension {} in GzipHandler".format(ext)
            )

    def exists(self, name):
        return name == self.name

    def list_files(self, sub_path):
        return [self.name]

    def open(self, name):
        if name != self.name:
            raise Exception("File not found")
        return gzip.open(self.path, "rb")

    def read(self, name):
        if name != self.name:
            raise Exception("File not found")
        return gzip.open(self.path, "rb").read()


class XzHandler(object):
    def __init__(self, path):
        self.path = path
        name, ext = os.path.splitext(os.path.basename(path))
        ext = ext.lower()
        if ext == ".xz":
            self.name = name
        else:
            raise Exception("Unexpected extension {} in XzHandler".format(ext))

    def exists(self, name):
        return name == self.name

    def list_files(self, sub_path):
        return [self.name]

    def open(self, name):
        if name != self.name:
            raise Exception("File not found")
        return lzma.open(self.path, "rb")

    def read(self, name):
        if name != self.name:
            raise Exception("File not found")
        return lzma.open(self.path, "rb").read()


class SevenZipHandler(object):
    def __init__(self, path):
        self.path = path
        self.zip = SevenZipFile(self.path, "r")

    def exists(self, name):
        try:
            self.zip.getinfo(name)
        except KeyError:
            return False
        else:
            return True

    def list_files(self, sub_path):
        if sub_path:
            return
        return self.zip.namelist()

    def open(self, name):
        data = self.zip.read(name)
        return io.BytesIO(data)

    def read(self, name):
        return self.zip.read(name)


class LhaHandler(object):
    def __init__(self, path):
        self.path = path
        self._lhafile = LhaFile(self.path, "r")
        self._nameMapping = {}
        for name in self._lhafile.NameToInfo:
            self._nameMapping[self.decode_name(name)] = name

    def decode_name(self, name):
        # Normalize both backslash and slash to forward slashes only
        name = name.replace("\\", "/")
        return name

    def encode_name(self, name):
        # Convert old backslashes (older Launcher/Windows) to slashes.
        name = name.replace("\\", "/")
        # Unescape old escapes.
        name = name.replace("%5c", "/")
        name = name.replace("%25", "%")
        # Legacy workaround for existing entries with incorrect escape.
        name = name.replace("%5f", "/")
        return self._nameMapping[name]

    def exists(self, name):
        # Looking up in NameToInfo isn't really necessary, since encode_name
        # by itself currently throws a KeyError if the name is not known.
        # However, to avoid bugs in the future if encode_name implementation
        # changes, we both normalize the name and look it up.
        try:
            return self.encode_name(name) in self._lhafile.NameToInfo
        except KeyError:
            return False

    def getinfo(self, name):
        # FIXME: Should instead add getinfo to LhaFile...
        return self._lhafile.NameToInfo[self.encode_name(name)]

    def list_files(self, sub_path):
        if sub_path:
            return
        for name in self._lhafile.namelist():
            # if name.endswith(str("/")):
            #     continue
            yield self.decode_name(name)

    def open(self, name):
        # LhaFile does not have open method
        data = self._lhafile.read(self.encode_name(name))
        return io.BytesIO(data)


class NullHandler(object):
    def __init__(self, path):
        self.path = path

    def list_files(self, _):
        return []

    def open(self, path):
        return filter_open(path)


def filter_open(path, stream=None):
    if stream is None:
        stream = open(path.rsplit("#?", 1)[0], "rb")
    if "#?" in path:
        if path.endswith("#?Filter=Skip(16)"):
            return SkipFilter(stream, 16)
        elif path.endswith("#?Filter=Skip(128)"):
            return SkipFilter(stream, 128)
        elif path.endswith("#?Filter=Skip(512)"):
            return SkipFilter(stream, 512)
        elif path.endswith("#?Filter=ByteSwapWords"):
            return ByteSwapWordsFilter(stream)
        else:
            raise Exception(
                "Unrecognized file filter: " + path.split("#?")[-1]
            )
    return stream


class SkipFilter:
    def __init__(self, stream, count):
        print("[ARCHIVE] Skip({}) filter for".format(count), stream)
        self.stream = stream
        self.count = count
        self.strip_left = count

    def read(self, n=-1):
        if self.strip_left:
            data = self.stream.read(self.strip_left)
            if len(data) == 0:
                # No data returned, end of stream
                return b""
            self.strip_left -= len(data)
        return self.stream.read(n)


class ByteSwapWordsFilter:
    def __init__(self, stream):
        print("[ARCHIVE] ByteSwapWords filter for", stream)
        self.stream = stream

    def read(self, n=-1):
        data = self.stream.read(n)
        # FIXME: Support odd length reads
        assert len(data) % 2 == 0
        io = BytesIO()
        for i in range(0, len(data), 2):
            io.write(data[i + 1 : i + 2])
            io.write(data[i : i + 1])
        io.seek(0)
        return io.getvalue()


class Archive(object):

    extensions = archive_extensions

    def __init__(self, path):
        self.path, self.sub_path = self.split_path(path)
        self._handler = None

    def join(self, base, *args):
        return os.path.join(base, *args)

    def dirname(self, path):
        return os.path.dirname(path)

    def split_path(self, path):
        print("[ARCHIVE] Split path", path)
        if "#/" in path:
            parts = path.rsplit("#/", 1)
            archive = parts[0]
            # if not archive[-1] in "/\\" and os.path.exists(archive):
            if os.path.isfile(archive):
                sub_path = parts[1] if len(parts) > 1 else ""
                return archive, sub_path

        parts = path.replace("\\", "/").split("/")
        for i, part in enumerate(parts):
            n, ext = os.path.splitext(part)
            ext = ext.lower()
            if ext in archive_extensions:
                # FIXME: should also check that it isn't a dir
                path = str(os.sep).join(parts[: i + 1])
                sub_path = str(os.sep).join(parts[i + 1 :])
                return path, sub_path
        return path, ""

    def get_handler(self):
        if self._handler is not None:
            return self._handler
        print("[ARCHIVE] get_handler", self.path)
        name, ext = os.path.splitext(self.path)
        ext = ext.lower()
        if ext == ".7z" and SevenZipFile is not None:
            self._handler = SevenZipHandler(self.path)
            return self._handler
        if ext in archive_extensions_gz:
            self._handler = GzipHandler(self.path)
            return self._handler
        if ext in archive_extensions_xz:
            self._handler = XzHandler(self.path)
            return self._handler
        try:
            self._handler = ZipHandler(self.path)
        except Exception as e:
            if ext == ".zip":
                print("[ARCHIVE]", repr(e))
            try:
                self._handler = LhaHandler(self.path)
            except Exception as e:
                if ext == ".lha":
                    print("[ARCHIVE]", repr(e))
                self._handler = NullHandler(self.path)
        return self._handler

    def list_files(self):
        result = []
        print("[ARCHIVE]", self.get_handler())
        for item in self.get_handler().list_files(self.sub_path):
            # result.append(os.path.join(self.path, item))
            result.append(self.path + "#/" + item)
        return result

    def exists(self, path):
        path, sub_path = self.split_path(path)
        # print(path, self.path)
        assert path == self.path
        if not sub_path:
            if "#?" in path:
                path = path.rsplit("#?", 1)[0]
            return os.path.exists(path)
        if "#?" in sub_path:
            sub_path = sub_path.rsplit("#?", 1)[0]
        return self.get_handler().exists(sub_path)

    def getinfo(self, path):
        path, sub_path = self.split_path(path)
        # print(path, self.path)
        assert path == self.path
        # if not sub_path:
        #     if "#?" in path:
        #         path = path.rsplit("#?", 1)[0]
        #     return os.path.exists(path)
        if "#?" in sub_path:
            sub_path = sub_path.rsplit("#?", 1)[0]
        return self.get_handler().getinfo(sub_path)

    def open(self, path):
        print("[Archive] Open", repr(path))
        path, sub_path = self.split_path(path)
        # print(path, self.path)
        assert path == self.path
        if not sub_path:
            return filter_open(path)
        if "#?" in sub_path:
            sub_path = sub_path.rsplit("#?", 1)[0]
        return filter_open(path, self.get_handler().open(sub_path))

    def copy(self, path, dest):
        ifo = self.open(path)
        with open(dest, "wb") as ofo:
            while True:
                data = ifo.read(65536)
                if not data:
                    break
                ofo.write(data)



import fsui.qt
from fsui.qt import Qt
from .widget_mixin import WidgetMixin
from urllib.parse import unquote_plus


class Label(fsui.qt.QLabel, WidgetMixin):
    def __init__(self, parent, label):
        fsui.qt.QLabel.__init__(self, label, parent.get_container())
        self.init_widget(parent)

        self.setTextFormat(fsui.qt.Qt.RichText)
        # self.setTextInteractionFlags(fsui.qt.Qt.TextBrowserInteraction)
        self.setTextInteractionFlags(
            Qt.TextSelectableByMouse | Qt.LinksAccessibleByMouse
        )
        self.setOpenExternalLinks(True)
        # self.setFocusPolicy(Qt.NoFocus)

        # FIXME: focusPolicy()
        # FIXME: make Label more plain, and rather make a InteractiveLabel
        # descendant or something like that

    def set_text_color(self, color):
        palette = self.palette()
        palette.setColor(self.foregroundRole(), color)
        self.setPalette(palette)

    def set_text(self, label):
        self.setText(label)


class PlainLabel(fsui.qt.QLabel, WidgetMixin):
    def __init__(self, parent, label):
        fsui.qt.QLabel.__init__(self, label, parent.get_container())
        self.init_widget(parent)

    def set_text(self, label):
        self.setText(label)


class URLLabel(Label):
    def __init__(self, parent, label, url):
        self._label = label
        self._url = url
        Label.__init__(self, parent, self._fix_label())
        # self.setFocusPolicy(Qt.StrongFocus)

    def set_text(self, label):
        self._label = label
        self.setText(self._fix_label())

    def set_url(self, url):
        self._url = url
        self.setText(self._fix_label())

    def _fix_label(self):
        url = unquote_plus(self._url)
        return '<a href="{0}">{1}</a>'.format(url, self._label)

    def get_min_height(self):
        # because the underline seems to be cut off otherwise...
        return Label.get_min_height(self) + 1


class MultiLineLabel(WidgetMixin):
    def __init__(self, parent, label, min_width=None):
        self._widget = fsui.qt.QLabel(label, parent.get_container())
        # Widget.__init__(self, parent)
        self.init_widget(parent)
        self._widget.setWordWrap(True)
        # self._widget.setFixedWidth(200)
        # self._widget.setFixedHeight(200)
        if min_width:
            self.set_min_width(min_width)

        self._widget.setTextFormat(fsui.qt.Qt.RichText)
        self._widget.setTextInteractionFlags(fsui.qt.Qt.TextBrowserInteraction)
        self._widget.setOpenExternalLinks(True)
        self._widget.setAlignment(fsui.qt.Qt.AlignLeft | fsui.qt.Qt.AlignTop)

    def set_text(self, label):
        self._widget.setText(label)

    # def set_fixed_width(self, width):

    # def set_min_width(self):
    #     pass

    def get_min_height(self):
        # + 1 because of url underlines
        if hasattr(self, "min_width"):
            if self.min_width:
                height = self._widget.heightForWidth(self.min_width) + 1
                if hasattr(self, "min_height"):
                    return max(self.min_height, height)
                return height
        return WidgetMixin.get_min_height(self) + 1



from fsgs.ogd.client import OGDClient
import fsui
from fsbc.application import app
from launcher.ui.widgets import CloseButton
from workspace.apps.refresh import RefreshWindow

# from workspace.shell import SimpleApplication
from launcher.res import gettext
from workspace.ui.theme import WorkspaceTheme


class LoginWindow(fsui.Window):
    @classmethod
    def open(cls, parent=None):
        return fsui.open_window_instance(cls, parent)

    def __init__(self, parent=None):
        print("LoginWindow, parent =", parent)
        super().__init__(parent, gettext("Log In to Your OpenRetro Account"))
        self.set_icon(fsui.Icon("password", "pkg:workspace"))
        self.theme = WorkspaceTheme.instance()
        self.layout = fsui.VerticalLayout()

        self.layout.set_padding(20, 20, 20, 20)

        heading_layout = fsui.HorizontalLayout()
        self.layout.add(heading_layout)
        heading_layout.add(
            fsui.ImageView(self, fsui.Image("workspace:res/48/password.png"))
        )
        heading_layout.add_spacer(20)
        heading_layout_2 = fsui.VerticalLayout()
        heading_layout.add(
            heading_layout_2, expand=True, fill=False, valign=0.5
        )
        heading_layout_2.add(
            fsui.HeadingLabel(
                self, gettext("Log In to Your OpenRetro Account")
            )
        )
        heading_layout_2.add_spacer(2)
        heading_layout_2.add(
            fsui.Label(
                self,
                gettext(
                    "Logging in will enable the online game database "
                    "and more"
                ),
            )
        )

        self.username_field = fsui.TextField(
            self, app.settings["database_email"].strip()
        )
        self.password_field = fsui.PasswordField(self)
        if self.username_field.get_text():
            self.password_field.focus()

        self.layout.add_spacer(20)
        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True)
        label = fsui.Label(self, gettext("E-mail:"))
        label.set_min_width(100)
        hori_layout.add(label)
        hori_layout.add_spacer(20)
        # self.username_field.select_all()
        self.username_field.changed.connect(self.on_text_field_changed)
        self.username_field.activated.connect(self.on_username_activated)
        hori_layout.add(self.username_field, expand=True)

        self.layout.add_spacer(10)
        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True)
        label = fsui.Label(self, gettext("Password:"))
        label.set_min_width(100)
        hori_layout.add(label)
        hori_layout.add_spacer(20)
        self.password_field.changed.connect(self.on_text_field_changed)
        self.password_field.activated.connect(self.on_password_activated)
        hori_layout.add(self.password_field, expand=True)

        self.layout.add_spacer(20)
        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True)
        label = fsui.Label(self, gettext("Don't have an account already?"))
        hori_layout.add(label)
        hori_layout.add_spacer(20)
        label = fsui.URLLabel(
            self,
            gettext("Create an account now"),
            "https://openretro.org/user/register?referrer=fs-uae-launcher",
        )
        hori_layout.add(label, expand=True)

        self.layout.add_spacer(6)
        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True)
        label = fsui.Label(self, gettext("Forgot your password?"))
        hori_layout.add(label)
        hori_layout.add_spacer(20)
        self.reset_label = fsui.URLLabel(
            self,
            gettext("Reset password via e-mail"),
            "https://openretro.org/user/reset?referrer=fs-uae-launcher",
        )
        hori_layout.add(self.reset_label, expand=True)

        self.layout.add_spacer(20)
        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True)
        self.created_label = fsui.Label(self, "")
        hori_layout.add(self.created_label, expand=True)
        hori_layout.add_spacer(20)

        self.login_button = fsui.Button(self, gettext("Log In"))
        self.login_button.disable()
        self.login_button.activated.connect(self.on_login_activated)
        hori_layout.add(self.login_button)

        if self.window.theme.has_close_buttons:
            self.close_button = CloseButton(self)
            hori_layout.add(self.close_button, fill=True, margin_left=10)

        self.set_size(self.layout.get_min_size())
        self.center_on_parent()

        if len(self.username_field.text()) == 0:
            self.username_field.focus()
        else:
            self.password_field.focus()

    def __del__(self):
        print("LoginWindow.__del__")

    def on_text_field_changed(self):
        email = self.username_field.get_text().strip()
        password = self.password_field.get_text().strip()
        self.login_button.enable(bool(email) and bool(password))
        self.reset_label.set_url(
            "https://openretro.org/user/reset?referrer=fs-uae-launcher"
            "&email={0}".format(email)
        )

    def on_username_activated(self):
        if self.username_field.get_text().strip():
            self.password_field.focus()

    def on_password_activated(self):
        self.on_login_activated()

    def on_login_activated(self):
        email = self.username_field.get_text().strip()
        password = self.password_field.get_text().strip()
        if not email or not password:
            return
        self.username_field.disable()
        self.password_field.disable()
        self.login_button.disable()

        task = OGDClient().login_task(email, password)
        # task.progressed.connect(self.progress)
        task.succeeded.connect(self.on_success)
        # task.failed.connect(fsui.error_function(gettext("Login Failed")))
        task.failed.connect(self.on_failure)
        task.start()

    def on_success(self):
        # center = self.get_window_center()
        # fsui.call_after(start_refresh_task, center)
        # shell_open("Workspace:Tools/Refresh", center=self.get_window_center())
        RefreshWindow.open(self.parent())
        self.close()

    def on_failure(self, message):
        fsui.show_error(message, parent=self.get_window())
        self.username_field.enable()
        self.password_field.enable()
        self.login_button.enable()
        self.password_field.select_all()
        self.password_field.focus()

    # def on_progress(self):
    #    print("on_progress")

    # def on_success(self):
    #     self.close()


# def start_refresh_task(center):
#     RefreshWindow.open()


# application = SimpleApplication(LoginWindow)



import fsui
from fsgs.Database import Database
from fsgs.ogd.client import OGDClient
from fsui import Image, Choice
from launcher.i18n import gettext
from launcher.launcher_config import LauncherConfig
from launcher.launcher_settings import LauncherSettings
from launcher.ui.ConfigGroup import ConfigGroup
from launcher.ui.ConfigurationsBrowser import ConfigurationsBrowser
from launcher.ui.GameListSelector import GameListSelector
from launcher.ui.behaviors.settingsbehavior import SettingsBehavior
from launcher.ui.skin import Skin
from launcher.ui.VariantsBrowser import VariantsBrowser
from launcher.ui.behaviors.configbehavior import ConfigBehavior
from launcher.ui.newbutton import NewButton


class ConfigurationsPanel(fsui.Panel):
    def __init__(self, parent):
        fsui.Panel.__init__(self, parent)
        Skin.set_background_color(self)
        self.layout = fsui.HorizontalLayout()

        vert_layout = fsui.VerticalLayout()
        self.layout.add(vert_layout, fill=True, expand=True)

        hor_layout = fsui.HorizontalLayout()
        vert_layout.add(hor_layout, fill=True)

        label_stand_in = fsui.Panel(self)
        tw, th = label_stand_in.measure_text("Games")
        label_stand_in.set_min_height(th)
        hor_layout.add(label_stand_in, margin_top=10, margin_bottom=10)

        hor_layout.add(NewButton(self), margin_left=10, margin_right=10)

        game_list_selector = GameListSelector(self)
        game_list_selector.set_min_width(250)
        game_list_selector.setMaximumWidth(250)
        game_list_selector.changed.connect(self.on_game_list_changed)
        hor_layout.add(game_list_selector, expand=False, margin_left=10)

        self.text_field = fsui.TextField(
            self, LauncherSettings.get("config_search")
        )
        self.text_field.on_changed = self.on_search_changed
        if VariantsBrowser.use_horizontal_layout():
            # window is big enough to use fixed size
            # self.text_field.set_min_width(210)
            self.text_field.set_min_width(229)
            hor_layout.add(
                self.text_field,
                expand=False,
                margin=10,
                margin_top=0,
                margin_bottom=0,
            )
        else:
            hor_layout.add(
                self.text_field,
                expand=True,
                margin=10,
                margin_top=0,
                margin_bottom=0,
            )

        # self.refresh_button = IconButton(self, "refresh_button.png")
        # self.refresh_button.set_tooltip(
        #     gettext("Refresh Game Configurations from Online Database"))
        # self.refresh_button.activated.connect(self.on_refresh_button)
        # hor_layout.add(
        #     self.refresh_button, margin=10, margin_top=0, margin_bottom=0)

        self.configurations_browser = ConfigurationsBrowser(self)

        vert_layout.add(
            self.configurations_browser, fill=True, expand=3, margin=10
        )

        self.variants_panel = fsui.Panel(self)
        vert_layout.add(
            self.variants_panel,
            fill=True,
            expand=False,
            margin=10,
            margin_top=20,
        )

        self.variants_panel.layout = fsui.HorizontalLayout()
        self.variants_browser = VariantsBrowser(self.variants_panel)
        # Do not use fill=True with the default OS X theme at least,
        # if you do the item will be rendered with the old Aqua look
        self.variants_panel.layout.add(
            self.variants_browser, fill=False, expand=True
        )

        # for rating in [1, 4, 5]:
        #     button = RatingButton(self.variants_panel, rating)
        #     self.variants_panel.layout.add(button, margin_left=5, fill=True)

        self.variants_panel.layout.add(
            RatingChoice(self.variants_panel), margin_left=5, fill=True
        )

        self.config_panel = fsui.Panel(self)
        vert_layout.add(
            self.config_panel,
            fill=True,
            expand=False,
            margin_bottom=10,
            margin_top=20,
        )

        self.config_panel.layout = fsui.VerticalLayout()
        self.config_group = ConfigGroup(self.config_panel, new_button=False)
        self.config_panel.layout.add(self.config_group, fill=True, expand=True)

        LauncherSettings.add_listener(self)
        self.on_setting("parent_uuid", LauncherSettings.get("parent_uuid"))

    def on_destroy(self):
        LauncherSettings.remove_listener(self)

    def on_setting(self, key, value):
        if key == "parent_uuid":
            self.variants_panel.show_or_hide(bool(value))
            self.config_panel.show_or_hide(not bool(value))
            self.layout.update()

    def on_verified_button(self):
        pass

    def on_favorite_button(self):
        pass

    def on_search_changed(self):
        text = self.text_field.get_text()
        LauncherSettings.set("config_search", text)

    def on_game_list_changed(self):
        print("ConfigurationsPanel.on_game_list_changed!")
        self.text_field.set_text("")


class RatingChoice(Choice):
    """Control allowing the user to set rating for a variant.

    The control disallows the use of cursor keys for directly changing
    the selected item, to avoid accidental ratings.
    """

    def __init__(self, parent):
        self.active_icon = 1
        super().__init__(parent, [], cursor_keys=False)
        with self.changed.inhibit:
            self.add_item(
                gettext("Rate Variant"), Image("launcher:res/16x16/bullet.png")
            )
            self.add_item(
                gettext("Best Variant"),
                Image("launcher:res/16x16/rating_fav_2.png"),
            )
            self.add_item(
                gettext("Good Variant"),
                Image("launcher:res/16x16/thumb_up_2.png"),
            )
            self.add_item(
                gettext("Bad Variant"),
                Image("launcher:res/16x16/thumb_down_2.png"),
            )
        ConfigBehavior(self, ["variant_uuid"])
        SettingsBehavior(self, ["__variant_rating"])

    def on_changed(self):
        variant_uuid = LauncherConfig.get("variant_uuid", "")
        if not variant_uuid:
            return
        rating = self.index_to_rating(self.get_index())
        self.set_rating_for_variant(variant_uuid, rating)

    @staticmethod
    def index_to_rating(index):
        return [0, 5, 4, 1][index]

    @staticmethod
    def rating_to_index(rating):
        return [0, 3, 3, 2, 2, 1][rating]

    @staticmethod
    def set_rating_for_variant(variant_uuid, rating):
        # FIXME: Do asynchronously, add to queue
        client = OGDClient()
        result = client.rate_variant(variant_uuid, like=rating)
        like_rating = result.get("like", 0)
        work_rating = result.get("work", 0)
        database = Database.instance()
        cursor = database.cursor()
        cursor.execute(
            "DELETE FROM rating WHERE game_uuid = ?", (variant_uuid,)
        )
        cursor.execute(
            "INSERT INTO rating (game_uuid, work_rating, like_rating) "
            "VALUES (?, ?, ?)",
            (variant_uuid, work_rating, like_rating),
        )
        database.commit()
        LauncherSettings.set("__variant_rating", str(like_rating))

    def on___variant_rating_setting(self, value):
        with self.changed.inhibit:
            try:
                rating = int(value)
            except ValueError:
                rating = 0
            self.set_index(self.rating_to_index(rating))

    def on_variant_uuid_config(self, value):
        self.set_enabled(value != "")



from .element import LightElement


class Spacer(LightElement):
    def __init__(self, size, size2=None):
        LightElement.__init__(self, None)
        self.width = size
        if size2 is None:
            self.height = size
        else:
            self.height = size2

    def get_min_size(self):
        return self.width, self.height

    def get_min_width(self):
        return self.width

    def get_min_height(self):
        return self.height



from fsui.qt.helpers import QParent
from fsui.qt.widget import Widget


class Adapter(Widget):
    def __init__(self, parent, widget):
        super().__init__(parent)
        self._widget = widget
        self._widget.setParent(QParent(parent))
        # self._widget.move(0, 2000)

        # def on_resize(self):
        #     self._widget.setGeometry(0, 0, self.width(), self.height())



import fsbc.desktop
import nose.tools


def test_mypy():
    # from nose.plugins.skip import SkipTest
    # raise SkipTest()
    import fstd.mypy

    fstd.mypy.check_module(fsbc.desktop.__name__)


def test_doctest():
    import doctest

    failure_count, test_count = doctest.testmod(fsbc.desktop)
    nose.tools.assert_equals(failure_count, 0)



import fsui
from launcher.i18n import gettext
from launcher.option import Option
from launcher.ui.IconButton import IconButton
from launcher.ui.behaviors.configbehavior import ConfigBehavior
from launcher.ui.behaviors.platformbehavior import AmigaShowBehavior
from launcher.ui.config.configpanel import ConfigPanel
from launcher.ui.config.configdialog import ConfigDialog


class AdditionalConfigPanel(ConfigPanel):
    def __init__(self, parent):
        super().__init__(parent)
        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True)
        heading_label = fsui.HeadingLabel(
            self, gettext("Additional Configuration")
        )
        hori_layout.add(heading_label, margin=10)
        hori_layout.add_spacer(0, expand=True)
        hori_layout.add(CustomConfigButton(self), margin_right=10)
        self.layout.add_spacer(0)

        amiga_panel = fsui.Panel(self)
        amiga_panel.layout = fsui.VerticalLayout()
        self.layout.add(amiga_panel, fill=True)
        self.add_amiga_option(Option.CPU, parent=amiga_panel)
        self.add_amiga_option(Option.JIT_COMPILER, parent=amiga_panel)
        self.add_amiga_option(Option.FLOPPY_DRIVE_SPEED, parent=amiga_panel)
        self.add_amiga_option(
            Option.FLOPPY_DRIVE_VOLUME_EMPTY, parent=amiga_panel
        )
        self.add_amiga_option(Option.FREEZER_CARTRIDGE, parent=amiga_panel)
        self.add_amiga_option(Option.DONGLE_TYPE, parent=amiga_panel)

    #     AmigaShowBehavior(amiga_panel)
    #     ConfigBehavior(self, [Option.PLATFORM])
    #
    # def on_platform_config(self, _):
    #     self.layout.update()


class CustomConfigButton(IconButton):
    def __init__(self, parent):
        super().__init__(parent, "16x16/settings.png")
        self.activated.connect(self.__activated)

    def __activated(self):
        ConfigDialog.run(self.window)



import time

from arcade.notification import Notification
from fsgs.input.device import Device
from fsgs.input.eventhelper import EventHelper
from fsgs.input.devicemanager import DeviceManager

CHECK_DELAY = 0.5


class EventListener:
    def __init__(self):
        self.helper = EventHelper()
        self.helper.init()
        self.device_names = {}
        self.device_manager = DeviceManager(None)
        self.remove_count = 0
        self.restart_after_add_count = 0
        self.reinit_time = 0.0
        self.check_time = 0.0
        self.check_list = None

    def get_next_event(self):
        # print(self.reinit_time - self.check_time)
        if self.reinit_time > self.check_time:
            delay = CHECK_DELAY
            if self.check_time == 0.0:
                # Force longer delay on startup to suppress false insert
                # notifications on the initial set of devices.
                delay = 2.5
            if time.time() - self.reinit_time > delay:
                self.check_for_device_changes()

        event = self.helper.get_next_event()
        if event is None:
            return None

        if self.helper.is_remove_event(event):
            print("[INPUT] Remove event:", event)
            # Restarting event helper so all devices are re-enumerated
            self.remove_count += 1
            self.reinit()

        if self.helper.is_add_event(event):
            if self.restart_after_add_count < self.remove_count:
                # New device added after a device has been removed. Restarting
                # device helper so device enumeration is consistent.
                # Enumeration must be the same as emulators will do by
                # themselves.
                self.restart_after_add_count += 1
                self.reinit()
                return

            count = self.device_names.get(event["name"], 0) + 1
            self.device_names[event["name"]] = count
            if count > 1:
                event["id"] = "{0} #{1}".format(event["name"], count)
            else:
                event["id"] = event["name"]
            # self.event_id = event["type"] + "-" + event["device"]
            self.device_manager.add_device_from_event(event)
            # self.notification((device.type, device.id))
            # self.check_for_device_changes()
            self.reinit_time = time.time()

        return event

    def reinit(self):
        self.device_manager.remove_all_devices()
        self.device_names = {}
        self.helper = EventHelper()
        self.helper.init()
        self.reinit_time = time.time()

    def check_for_device_changes(self):
        print("[INPUT] Checking for device change notifications")
        print(time.time(), self.reinit_time, self.check_time)
        new_list = []
        for device in self.device_manager.get_devices():
            new_list.append((device.type, device.id))
        old_list = self.check_list
        if old_list is not None:
            print("[INPUT] Old:", old_list)
            print("[INPUT] New:", new_list)
            old_set = set(old_list)
            new_set = set(new_list)
            for item in old_list:
                if item not in new_set:
                    self.notification(item, "Removed")
            for item in new_list:
                if item not in old_set:
                    self.notification(item, "Added")
        self.check_list = new_list
        self.check_time = time.time()

    def notification(self, item, action):
        if item[0] == Device.TYPE_JOYSTICK:
            text = "Joystick {}:\n{}".format(action, item[1])
        elif item[0] == Device.TYPE_KEYBOARD:
            text = "Keyboard {}:\n{}".format(action, item[1])
        elif item[0] == Device.TYPE_MOUSE:
            text = "Mouse {}:\n{}".format(action, item[1])
        else:
            text = "Device {}:\n{}".format(action, item[1])
        Notification(text)



from __future__ import absolute_import
from __future__ import print_function

# a block device defines a set of blocks used by a file system
from .DiskGeometry import DiskGeometry

class BlockDevice:
  def _set_geometry(self, cyls=80, heads=2, sectors=11, block_bytes=512, reserved=2, bootblocks=2):
    self.cyls = cyls
    self.heads = heads
    self.sectors = sectors
    self.block_bytes = block_bytes
    self.reserved = reserved
    self.bootblocks = bootblocks
    # derived values
    self.num_tracks = self.cyls * self.heads
    self.num_blocks = self.num_tracks * self.sectors
    self.num_bytes = self.num_blocks * self.block_bytes
    self.block_longs = self.block_bytes / 4
    self.num_longs = self.num_blocks * self.block_longs
  
  def dump(self):
    print("cylinders:  ", self.cyls)
    print("heads:      ", self.heads)
    print("sectors:    ", self.sectors)
    print("block_bytes:", self.block_bytes)
    print("reserved:   ", self.reserved)
    print("bootblocks: ", self.bootblocks)

  def _blk_to_offset(self, blk_num):
    return self.block_bytes * blk_num
  
  # ----- API -----
  def create(self, **args):
    pass
  def open(self):
    pass
  def close(self):
    pass
  def flush(self):
    pass
  def read_block(self, blk_num):
    pass
  def write_block(self, blk_num, data):
    pass
  def get_geometry(self):
    return DiskGeometry(self.cyls, self.heads, self.sectors)
  def get_chs_str(self):
    return "chs=%d,%d,%d" % (self.cyls, self.heads, self.sectors)
  def get_chs_dict(self):
    return { 'chs' : "%d,%d,%d" % (self.cyls, self.heads, self.sectors) }



import traceback

import fsui
from fsgs.option import Option
from fsgs.platform import Platform
from launcher.i18n import gettext
from launcher.launcher_config import LauncherConfig
from launcher.option import options
from launcher.settings.fullscreenmodebutton import FullscreenModeButton
from launcher.settings.monitorbutton import MonitorButton
from launcher.settings.option_ui import OptionUI
from launcher.settings.platformsettingsdialog import PlatformSettingsDialog
from launcher.settings.settings_dialog import SettingsDialog
from launcher.settings.videosynccheckbox import VideoSyncCheckBox
from launcher.ui.IconButton import IconButton
from launcher.ui.behaviors.configbehavior import ConfigBehavior
from launcher.ui.behaviors.platformbehavior import (
    PlatformShowBehavior,
    AMIGA_PLATFORMS,
    PlatformEnableBehavior,
)
from launcher.ui.behaviors.settingsbehavior import SettingsBehavior

MEDNAFEN = [
    Platform.SNES,
    Platform.NES,
    Platform.GB,
    Platform.GBA,
    Platform.GBC,
    Platform.PSX,
    Platform.SMD,
    Platform.SMS,
    Platform.TG16,
    Platform.TGCD,
]
SCALING = [Platform.C64] + MEDNAFEN
STRETCHING = [Platform.C64, Platform.DOS, Platform.ZXS] + MEDNAFEN
EFFECTS = [Platform.C64, Platform.DOS, Platform.ZXS] + MEDNAFEN
BORDER = [Platform.C64, Platform.ZXS]
# BORDER += MEDNAFEN
SMOOTHING = [] + MEDNAFEN
CROPPING = [Platform.ZXS]

SCALING += AMIGA_PLATFORMS
STRETCHING += AMIGA_PLATFORMS
STRETCHING += [Platform.ARCADE, Platform.NEOGEO]

BEZEL = (
    AMIGA_PLATFORMS
    + MEDNAFEN
    + [
        Platform.ARCADE,
        Platform.NEOGEO,
        Platform.ZXS,
        Platform.DOS,
        Platform.C64,
    ]
)
CHEATS = MEDNAFEN + [Platform.ARCADE, Platform.NEOGEO]


class QuickSettingsPanel(fsui.Panel):
    def __init__(self, parent, fsgc):
        self.fsgc = fsgc
        super().__init__(parent)
        self.layout = fsui.VerticalLayout()

        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True)
        heading_label = fsui.HeadingLabel(self, gettext("Settings"))
        hori_layout.add(heading_label, margin=10)
        hori_layout.add_spacer(0, expand=True)
        settings_button = IconButton(self, "16x16/more.png")
        settings_button.activated.connect(self.on_settings_button)
        hori_layout.add(settings_button, margin_right=10)
        self.layout.add_spacer(0)

        # button = fsui.Button(self, "Platform Settings")
        # button.activated.connect(self.on_platform_settings_button)
        # self.layout.add(button, margin=10, fill=True)

        # self.add_option(Option.KEEP_ASPECT, text=gettext("Keep Aspect"),
        #                 platforms=AMIGA_PLATFORMS)
        self.add_option(Option.SCALE, text=None, enable=SCALING)
        self.add_option(Option.STRETCH, text=None, enable=STRETCHING)
        self.add_option(Option.BEZEL, text=None, enable=BEZEL)

        self.add_option(Option.EFFECT, text=None, enable=EFFECTS)

        self.add_option(Option.ZOOM, text=None, platforms=AMIGA_PLATFORMS)
        self.add_option(Option.BORDER, text=None, platforms=BORDER)

        # self.add_option(Option.SMOOTHING, text=None, platforms=SMOOTHING)

        # self.add_option(Option.CROP, text=None, platforms=CROPPING)

        # if fsgc.settings[Option.DEVELOPER_MODE] == "1":
        #     pass
        #     # self.add_option(Option.ZXS_DRIVER, [Platform.ZXS])
        #     # self.add_option(Option.DOS_EMULATOR, [Platform.DOS])
        # self.add_option(Option.NES_EMULATOR, [Platform.NES])

        self.add_option(Option.AUTO_LOAD, [Platform.DOS, Platform.ZXS])
        self.add_option(Option.AUTO_QUIT, [Platform.DOS])
        self.add_option(Option.TURBO_LOAD, [Platform.ZXS])

        # self.add_option(Option.FRAME, text=None, platforms=BEZEL)
        # self.add_option(Option.BEZEL, text=None, platforms=BEZEL)

        # self.add_option(Option.CHEATS, platforms=CHEATS)

        quick_settings = fsgc.settings[Option.QUICK_SETTINGS_OPTIONS]
        for option in quick_settings.split(","):
            option = option.strip().lower()
            if "[" in option:
                # For future use of e.g.:
                # option1[platform1,platform2],option2[platform,...]
                option, platforms = option.split("[", 1)
            else:
                platforms = []
            if option in options:
                try:
                    self.add_option(option)
                except Exception:
                    print("Error adding quick setting")
                    traceback.print_exc()

        self.layout.add_spacer(expand=True)

        hori_layout = fsui.HorizontalLayout()
        hori_layout.add_spacer(expand=True)
        self.platform_settings_button = fsui.Button(self, "Platform Settings")
        self.platform_settings_button.activated.connect(
            self.on_platform_settings_button
        )
        hori_layout.add(self.platform_settings_button, margin=10)
        self.layout.add(hori_layout, fill=True)

        hori_layout = fsui.HorizontalLayout()
        hori_layout.add_spacer(expand=True)
        self.video_sync_checkbox = VideoSyncCheckBox(self)
        hori_layout.add(self.video_sync_checkbox, margin_right=10)
        self.layout.add(hori_layout, fill=True)
        self.monitor_button = MonitorButton(self)
        hori_layout.add(self.monitor_button, fill=True, margin_right=10)
        if False:
            self.fullscreen_mode_button = FullscreenModeButton(self)
            hori_layout.add(
                self.fullscreen_mode_button, fill=True, margin_right=10
            )
        self.layout.add_spacer(10)

        ConfigBehavior(self, [Option.PLATFORM])
        SettingsBehavior(self, [Option.G_SYNC])

    def add_option(self, option, platforms=None, enable=None, text=""):
        panel = fsui.Panel(self)
        panel.layout = fsui.VerticalLayout()
        panel.layout.add(
            OptionUI.create_group(
                panel, option, text, thin=True, help_button=False
            ),
            fill=True,
        )
        self.layout.add(panel, fill=True, margin=10)
        if platforms:
            PlatformShowBehavior(panel, platforms)
        elif enable:
            PlatformEnableBehavior(panel, enable)

    def on_platform_config(self, value):
        self.layout.update()
        self.platform_settings_button.enable(
            len(PlatformSettingsDialog.option_list_for_platform(value)) > 0
        )

    def on_g_sync_setting(self, value):
        self.video_sync_checkbox.enable(value != "1")

    def on_settings_button(self):
        SettingsDialog.open(self.window)

    def on_platform_settings_button(self):
        platform = LauncherConfig.get("platform")
        if platform:
            PlatformSettingsDialog.open(self.window, platform)

    def get_min_height(self):
        # Because we add a lot of controls, force min size to 0 to avoid
        # this control reporting too large min height at startup.
        return 0



from fsui.qt import QColor


class BaseColor(object):
    def mix(self, color, opacity=0.5):
        # print("mix", color)
        """Mixes this color with another color and returns the result.

        Arguments:
        color -- The overlay color to mix in (Color or wx.Colour)
        opacity -- The opacity of the overlay color in the range [0.0, 1.0]

        Returns a reference to self.
        """
        assert 0.0 <= opacity <= 1.0, "Invalid opacity"
        iopacity = 1 - opacity
        # return wx.Colour(
        self.set_components(
            int(self[0] * iopacity + color[0] * opacity),
            int(self[1] * iopacity + color[1] * opacity),
            int(self[2] * iopacity + color[2] * opacity),
        )
        return self


class Color(QColor, BaseColor):
    def __init__(self, *args):
        QColor.__init__(self)

        if len(args) == 1:
            c = args[0]
            if isinstance(c, QColor):
                self.setRgb(c.red(), c.green(), c.blue(), c.alpha())
            else:
                if len(c) == 3:
                    r, g, b = c
                    self.setRgb(r, g, b)
                else:
                    r, g, b, a = c
                    self.setRgba(r, g, b, a)
        elif len(args) == 3:
            self.setRgb(*args)
        elif len(args) == 4:
            self.setRgb(args[0], args[1], args[2])
            self.setAlpha(args[3])
        else:
            raise TypeError("Color object is not initialized")

    @staticmethod
    def mix_colors(basecolor, overlaycolor, opacity=0.5):
        c = Color(basecolor)
        return c.mix(overlaycolor, opacity)

    def set_components(self, r, g, b, a=255):
        self.setRgb(r, g, b)
        self.setAlpha(a)

    def __getitem__(self, index):
        if index == 0:
            return self.red()
        if index == 1:
            return self.green()
        if index == 2:
            return self.blue()
        raise IndexError("Invalid color component")

    def to_hex(self):
        if self.alpha() != 255:
            return "#{:02x}{:02x}{:02x}{:02x}".format(
                self.red(), self.green(), self.blue(), self.alpha()
            )
        return "#{:02x}{:02x}{:02x}".format(
            self.red(), self.green(), self.blue()
        )

    def to_hsl(self):
        return HSLColor.from_rgb(self.red(), self.green(), self.blue())

    def lighten(self, amount=0.05):
        c = self.to_hsl().lighten(amount).to_rgb()
        self.set_components(*c)
        return self

    def darken(self, amount=0.05):
        c = self.to_hsl().darken(amount).to_rgb()
        self.set_components(*c)
        return self

    # def mix(self, color, opacity=0.5):
    #     """Mixes this color with another color and returns the result.
    #
    #     Arguments:
    #     color -- The overlay color to mix in (Color or wx.Colour)
    #     opacity -- The opacity of the overlay color in the range [0.0, 1.0]
    #
    #     Returns a reference to self.
    #     """
    #     assert opacity >= 0.0 and opacity <= 1.0, "Invalid opacity"
    #     iopacity = 1 - opacity
    #     #return wx.Colour(
    #     self.Set(
    #             int(self.Red() * iopacity + color.Red() * opacity),
    #             int(self.Green() * iopacity + color.Green() * opacity),
    #             int(self.Blue() * iopacity + color.Blue() * opacity))
    #     return self

    def invert(self):
        self.Set(255 - self.Red(), 255 - self.Green(), 255 - self.Blue())
        return self

    def inverted(self):
        return self.copy().invert()

    def copy(self):
        return Color(*list(self))

    def complement(self):
        r, g, b = self.Red(), self.Green(), self.Blue()
        baseval = max(r, max(g, b)) + min(r, min(g, b))
        self.Set(
            baseval - self.Red(), baseval - self.Green(), baseval - self.Blue()
        )
        return self

    def complemented(self):
        return self.copy().complement()

    @property
    def intensity(self):
        return (self.Red() + self.Blue() + self.Green()) // 3


class HSLColor(BaseColor):
    def __init__(self):
        self.h = 0.0
        self.s = 0.0
        self.l = 0.0

    @staticmethod
    def from_rgb(r, g, b):
        r /= 255
        g /= 255
        b /= 255
        c = HSLColor()
        ma = max(r, max(g, b))
        mi = min(r, min(g, b))
        if ma == mi:
            c.h = 0
        elif ma == r:
            c.h = (60 * (g - b) / (ma - mi)) % 360
        elif ma == g:
            c.h = (60 * (b - r) / (ma - mi)) + 120
        else:
            c.h = (60 * (r - g) / (ma - mi)) + 240

        c.l = (ma + mi) / 2

        if ma == mi:
            c.s = 0
        elif c.l <= 0.5:
            c.s = (ma - mi) / (2 * c.l)
        else:  # c.l > 0.5
            c.s = (ma - mi) / (2 - 2 * c.l)
        return c

    def darken(self, amount=0.05):
        self.l = max(0.0, self.l - amount)
        return self

    def lighten(self, amount=0.05):
        self.l = min(1.0, self.l + amount)
        return self

    def to_rgb(self):

        if self.l < 0.5:
            q = self.l * (1 + self.s)
        else:  # c.l >= 0.5
            q = self.l + self.s - (self.l * self.s)
        p = 2 * self.l - q
        hk = self.h / 360

        # t = [tr, tg, tb]
        t = [hk + 1 / 3, hk, hk - 1 / 3]

        for i in range(3):
            if t[i] < 0.0:
                t[i] += 1.0
            elif t[1] > 1.0:
                t[i] -= 1.0

        for i in range(3):
            if t[i] < 1 / 6:
                t[i] = int(round(255 * (p + ((q - p) * 6 * t[i]))))
            elif 1 / 6 <= t[i] < 1 / 2:
                t[i] = int(round(255 * q))
            elif 1 / 2 <= t[i] < 2 / 3:
                t[i] = int(round(255 * (p + ((q - p) * 6 * (2 / 3 - t[i])))))
            else:
                t[i] = int(round(255 * p))
        return Color(t)

    def __getitem__(self, index):
        if index == 0:
            return int(round(self.h))
        if index == 1:
            return int(round(self.s * 100))
        if index == 2:
            return int(round(self.l * 100))
        raise IndexError("Invalid color component")

    def set_components(self, h, s, l):
        self.h = h
        self.s = s
        self.l = l



import os
import sys
import threading
import traceback
from .signal import Signal

main_thread_id = threading.current_thread().ident
already_run = False


def is_running_in_debugger():
    return bool(sys.gettrace())


def enable_exception_handler(force=False):
    global already_run
    if already_run:
        return
    already_run = True

    if not force and is_running_in_debugger():
        print(
            "running in debugger, not enabling exception handler " "after all."
        )
        _let_thread_exceptions_be_unhandled()
        return False

    # print("enabling except hook")
    sys.excepthook = _handle_exception
    _enable_thread_exception_handler()
    return True


def _let_thread_exceptions_be_unhandled():
    """Replace method in Thread class to let exceptions be picked up by
    the debugger. This is done by replacing the bootstrap function,
    which normally catches all exceptions."""

    print("let thread exceptions be unhandled (for debugger)")
    if sys.version.startswith("2.7"):
        threading.Thread._Thread__bootstrap = _thread_bootstrap_2_7
    else:
        print(
            "WARNING: no Thread bootstrap replacement for this "
            "python version. The debugger will not break on unhandled"
            "exceptions in threads other than the main thread."
        )


def _enable_thread_exception_handler():
    # print("enable tread exception handler")

    def run():
        self = threading.current_thread()
        try:
            self._run_()
        except:
            _handle_exception(*sys.exc_info())

    def set_ident(self):
        # self = threading.current_thread()
        self._run_ = self.run
        self.run = run
        self._set_ident_()

    # noinspection PyProtectedMember
    threading.Thread._set_ident_ = threading.Thread._set_ident
    threading.Thread._set_ident = set_ident


def _handle_exception(exc_type, exc_value, exc_traceback):
    if issubclass(exc_type, KeyboardInterrupt):
        print("handle_exception: KeyboardInterrupt caught")
        sys.exit(1)

    # FIXME: PYTHON3
    tb = traceback.extract_tb(exc_traceback)
    # print(exc_type, exc_value, tb)
    filename, line, function, dummy = tb.pop()
    try:
        filename = filename.decode(sys.getfilesystemencoding())
    except:
        pass
    try:
        filename = str(filename)
    except:
        pass
    filename = os.path.basename(filename)
    thread = threading.currentThread()
    error_id = "{0}:{1}:{2}:{3}".format(
        exc_type.__name__, filename, function, line
    )

    # QtGui.QMessageBox.critical(None, "Error",
    # description = "<html>A critical error has occurred.<br/> "
    #                            + "<b>%s</b><br/><br/>" % error
    #                            + "It occurred at <b>line %d</b> of file "
    #                              "<b>%s</b>.<br/>" % (
    #                            line, filename)
    #                            + "</html>")
    #

    backtrace_string = "".join(
        traceback.format_exception(exc_type, exc_value, exc_traceback)
    )

    message = "\nUnhandled exception detected in thread {0}:\n  {1}\n".format(
        thread.getName(), error_id
    )

    print(message)
    print(backtrace_string)

    print(message, file=sys.stderr)
    print(backtrace_string, file=sys.stderr)

    if thread.ident == main_thread_id:
        try:
            # FIXME
            # import wx
            # app = wx.GetApp()
            # if app is not None:
            #     wx.MessageBox(message + "\n" + backtrace_string)
            #     if app.IsMainLoopRunning():
            #         return
            #     print("The application will now close due to the above error")
            #     print("calling app.Exit")
            #     app.Exit()

            Signal("quit").notify()
        except Exception as e:
            print(repr(e))

        # sys.exit(1)


class AdditionalInfo(object):
    pass


def _thread_bootstrap_2_7(self):
    """This is a replacement "method" for the Thread class in Python 2.7,
    designed to let an exception fall through to the debugger."""

    # noinspection PyProtectedMember
    # noinspection PyUnresolvedReferences
    from threading import _active_limbo_lock, _active, _limbo, _trace_hook

    # noinspection PyProtectedMember
    # noinspection PyUnresolvedReferences
    from threading import _profile_hook, _sys, _get_ident

    try:
        self._set_ident()
        self._Thread__started.set()
        with _active_limbo_lock:
            _active[self._Thread__ident] = self
            del _limbo[self]
        if __debug__:
            self._note("%s.__bootstrap(): thread started", self)

        # if _trace_hook:
        #     self._note("%s.__bootstrap(): registering trace hook", self)
        #     _sys.settrace(_trace_hook)
        if _profile_hook:
            self._note("%s.__bootstrap(): registering profile hook", self)
            _sys.setprofile(_profile_hook)

        try:
            self.run()
        except SystemExit:
            if __debug__:
                self._note("%s.__bootstrap(): raised SystemExit", self)
        else:
            if __debug__:
                self._note("%s.__bootstrap(): normal return", self)
        finally:
            # Prevent a race in
            # test_threading.test_no_refcycle_through_target when
            # the exception keeps the target alive past when we
            # assert that it's dead.
            self._Thread__exc_clear()
    finally:
        with _active_limbo_lock:
            self._Thread__stop()
            try:
                # We don't call self.__delete() because it also
                # grabs _active_limbo_lock.
                del _active[_get_ident()]
            except:
                pass



from typing import Tuple
from fsui.qt import Qt, QPoint, QRect, QPen, QBrush, QLinearGradient, QColor
from fsui.qt import QFont
from .Color import Color


class Font(object):
    def __init__(self, name=None, size=None, font=None):
        if font is not None:
            self.font = font
        else:
            assert name
            self.font = QFont(name)
            if size:
                self.font.setPixelSize(size)

    def qfont(self):
        return self.font

    def set_bold(self, bold=True):
        self.font.setBold(bold)

    # def size(self):
    #     return self.font.pixelSize()
    #
    # def set_size(self, size):
    #     self.font.setPixelSize(size)

    def adjust_size(self, increment=1):
        size = self.font.pointSize()
        print("pontSize is", size)
        self.font.setPointSize(size + increment)

    def increase_size(self, increment=1):
        size = self.font.pointSize()
        print("pontSize is", size)
        self.font.setPointSize(size + increment)


class DrawingContext(object):
    def __init__(self, painter):
        self.qpainter = painter
        self.text_color = Color(0, 0, 0)

    def clear(self, color=None):
        # assuming here that the operations are clipped anyway, so just
        # using an larger-than-expected size
        rect = QRect(0, 0, 3000, 2000)
        if color is not None:
            self.qpainter.fillRect(rect, QBrush(color))
        else:
            self.qpainter.eraseRect(rect)

    def get_font(self):
        return Font(self.qpainter.font())

    def set_font(self, font):
        self.qpainter.setFont(font.font)

    def draw_text(self, text: str, x: int, y: int) -> None:
        # self.qpainter.drawText(QPoint(x, y), text)
        self.qpainter.setPen(QPen(self.text_color))
        self.qpainter.drawText(
            x, y, 10000, 1000, Qt.AlignLeft | Qt.AlignTop, text
        )

    def measure_text(self, text: str) -> Tuple[int, int]:
        # return self.dc.GetTextExtent(text)
        # return (10, 10)
        rect = self.qpainter.boundingRect(
            0, 0, 10000, 1000, Qt.AlignLeft | Qt.AlignTop, text
        )
        return rect.width(), rect.height()

    # def set_color(self, color):
    #     self.dc.SetPen(wx.Pen(color))
    #     #self.gc.SetPen(self.gc.CreatePen(wx.Pen(color)))

    def get_text_color(self):
        # return Color(self.dc.GetTextForeground())
        # pass
        # return Color(self.qpainter.pen().color())
        return self.text_color

    def set_text_color(self, color):
        # self.dc.SetTextForeground(color)
        # self.qpainter.setPen(QPen(color))
        self.text_color = Color(color)

    def draw_line(self, x1, y1, x2, y2, c):
        if x1 == x2:  # Line is vertical
            if y1 > y2:
                y1, y2 = y2, y1
            x2 += 1
        elif y1 == y2:  # Line is horizontal
            if x1 > x2:
                x1, x2 = x2, x1
            y2 += 1
        else:
            # Line is neither horizontal nor vertical, not supported right now
            return
        self.draw_rectangle(x1, y1, x2 - x1, y2 - y1, c)

    def draw_image(self, image, x, y):
        # self.dc.DrawBitmap(image.bitmap, x, y, True)
        self.qpainter.drawImage(QPoint(x, y), image.qimage)

    def draw_vertical_gradient(self, x, y, w, h, c1, c2):
        gradient = QLinearGradient(0, 0, 0, h)
        gradient.setColorAt(0.0, c1)
        gradient.setColorAt(1.0, c2)
        self.qpainter.setPen(QPen(QColor(0, 0, 0, 0)))
        self.qpainter.setBrush(QBrush(gradient))
        self.qpainter.drawRect(x, y, w, h)

    def draw_rectangle(self, x, y, w, h, c):
        self.qpainter.setPen(QPen(QColor(0, 0, 0, 0)))
        self.qpainter.setBrush(QBrush(c))
        self.qpainter.drawRect(x, y, w, h)

    def rounded_rectangle(
        self, x, y, w, h, c, radius, border_color=None, border_width=1.0
    ):
        if border_color is not None:
            self.qpainter.setPen(QPen(border_color, border_width))
        else:
            self.qpainter.setPen(QPen(QColor(0, 0, 0, 0)))
        self.qpainter.setBrush(QBrush(c))
        self.qpainter.drawRoundedRect(x, y, w, h, radius, radius)
